from motor.motor_asyncio import AsyncIOMotorClient

MONGO_URI = "mongodb://127.0.0.1:27017"


class Database:
    """
    A simple database object to house the Motor client.

    Attributes:
        client: An AsyncIOMotorClient for managing Scooze's MongoDB connection.
    """

    client: AsyncIOMotorClient = None


db = Database()


async def mongo_connect():
    """
    Connect to the database client to MongoDB.
    """

    db.client = AsyncIOMotorClient(MONGO_URI)


async def mongo_close():
    """
    Close the database client's MongoDB connection.
    """

    db.client.close()



import subprocess

from cleo.commands.command import Command


class TeardownLocalCommand(Command):
    name = "teardown local"
    description = "Teardown local Mongo database."

    def handle(self):
        # TODO(#201) - Add local MongoDB built-in support
        self.line(
            "Usage: <fg=cyan>scooze teardown docker</> or <fg=cyan>scooze teardown local</>. Local support coming soon."
        )



import uvicorn
from cleo.commands.command import Command


class RunCommand(Command):
    name = "run"
    description = "Run the Swagger UI/ReDocs."

    def handle(self):
        # TODO(6): Replace localhost with wherever we're hosting
        uvicorn.run("scooze.main:app", host="127.0.0.1", port=8000, reload=True)



from typing import Annotated, Any, TypeAlias

from bson import ObjectId as BsonObjectId
from pydantic import BaseModel, ConfigDict, GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import CoreSchema, core_schema

# region Private Utility Functions


def _to_lower_camel(string: str) -> str:
    words = string.split("_")

    if len(words) > 1:
        upper_camel = "".join(word.capitalize() for word in words)
        return upper_camel[0].lower() + upper_camel[1:]

    return words[0]


# endregion


# region Public Utility Classes


class ScoozeBaseModel(BaseModel, validate_assignment=True):
    """
    A simple base model class to support models in scooze.
    """

    model_config = ConfigDict(
        alias_generator=_to_lower_camel,
        arbitrary_types_allowed=True,
        populate_by_name=True,
    )


# Solution to BSON/MongoDB ObjectId issue, provided by Pydantic author:
# https://stackoverflow.com/a/76719893
class ObjectIdPydanticAnnotation:
    @classmethod
    def validate_object_id(cls, v: Any, handler) -> BsonObjectId:
        if isinstance(v, BsonObjectId):
            return v

        s = handler(v)
        if BsonObjectId.is_valid(s):
            return BsonObjectId(s)
        else:
            raise ValueError("Invalid ObjectId")

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type, _handler) -> core_schema.CoreSchema:
        assert source_type is BsonObjectId
        return core_schema.no_info_wrap_validator_function(
            cls.validate_object_id,
            core_schema.str_schema(),
            serialization=core_schema.to_string_ser_schema(),
        )

    @classmethod
    def __get_pydantic_json_schema__(cls, _core_schema: CoreSchema, handler: GetJsonSchemaHandler) -> JsonSchemaValue:
        return handler(core_schema.str_schema())


ObjectIdT: TypeAlias = Annotated[BsonObjectId, ObjectIdPydanticAnnotation]

# endregion



import scooze.database.deck as db
from fastapi import APIRouter
from fastapi.responses import JSONResponse
from scooze.models.deck import DeckModelIn

router = APIRouter(
    prefix="/deck",
    tags=["deck"],
    responses={404: {"description": "Deck Not Found"}},
)


@router.get("/", summary="Get a deck at random")
async def deck_root():
    decks = await db.get_decks_random(limit=1)
    if decks is not None:
        deck = decks[0]
        return JSONResponse(deck.model_dump(mode="json"), status_code=200)


# Create


@router.post("/add", summary="Create a new deck")
async def add_deck(deck: DeckModelIn):
    new_deck = await db.add_deck(deck=deck)
    if new_deck is not None:
        return JSONResponse(new_deck.model_dump(mode="json"), status_code=200)
    else:
        return JSONResponse({"message": f"Failed to create a new deck."}, status_code=400)


# Read


@router.get("/id/{deck_id}", summary="Get a deck by ID")
async def get_deck_by_id(deck_id: str):
    """
    Get the deck with the given scooze ID.

    - **deck_id** - the scooze ID of the deck to get
    """

    deck = await db.get_deck_by_property(property_name="_id", value=deck_id)
    if deck is not None:
        return JSONResponse(deck.model_dump(mode="json"), status_code=200)
    else:
        return JSONResponse({"message": f"Deck with id {deck_id} not found."}, status_code=404)


# Update


@router.patch("/update/{deck_id}", summary="Update an existing deck")
async def update_deck(deck_id: str, deck: DeckModelIn):
    """
    Update an existing deck with the given scooze ID and payload.

    Fields will be updated according to the given payload. If a field is not
    present in the payload, it will not be updated.

    - **deck_id** - the scooze ID of the deck to update
    """

    updated_deck = await db.update_deck(id=deck_id, deck=deck)

    if updated_deck is not None:
        return JSONResponse(updated_deck.model_dump(mode="json"), status_code=200)
    else:
        return JSONResponse({"message": f"Deck with id {deck_id} not found."}, status_code=404)


# Delete


@router.delete("/delete/{deck_id}", summary="Delete an existing deck")
async def delete_deck_by_id(deck_id: str):
    """
    Delete an existing deck with the given scooze ID.

    - **deck_id** - the scooze ID of the deck to delete
    """

    deleted_deck = await db.delete_deck(id=deck_id)

    if deleted_deck is not None:
        return JSONResponse({"message": f"Deck with id {deck_id} deleted."}, status_code=200)
    else:
        return JSONResponse({"message": f"Deck with id {deck_id} not deleted."}, status_code=400)



import os
from contextlib import asynccontextmanager

from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from scooze.database.mongo import mongo_close, mongo_connect
from scooze.routers.card import router as CardRouter
from scooze.routers.cards import router as CardsRouter
from scooze.routers.deck import router as DeckRouter
from scooze.routers.decks import router as DecksRouter


# Startup/shutdown
@asynccontextmanager
async def lifespan(app: FastAPI):
    # Setup Mongo
    await mongo_connect()

    # Yield to the app
    yield

    # Mongo teardown
    await mongo_close()


app = FastAPI(
    title="scooze",
    summary="REST API for interacting with MongoDB for Magic: the Gathering tournaments, decklists, and cards.",
    lifespan=lifespan,
)


# Router inclusion
app.include_router(CardRouter)
app.include_router(CardsRouter)
app.include_router(DeckRouter)
app.include_router(DecksRouter)

# Mount index.html
app_dir = os.path.dirname(__file__)
static_dir = os.path.join(app_dir, "static/")
app.mount("/", StaticFiles(directory=static_dir, html=True), name="static")



from typing import Any

import scooze.database.core as db_core
from bson import ObjectId
from scooze.catalogs import DbCollection
from scooze.models.deck import DeckModelIn, DeckModelOut

# region Deck


async def add_deck(deck: DeckModelIn) -> DeckModelOut:
    """
    Adds a deck to the database.

    Args:
        deck: The deck to insert.

    Returns:
        The deck that was inserted, or None if it was unable.
    """

    new_deck = await db_core.insert_document(
        DbCollection.DECKS,
        deck.model_dump(
            mode="json",
            by_alias=True,
        ),
    )
    if new_deck is not None:
        return DeckModelOut.model_validate(new_deck)


async def get_deck_by_property(property_name: str, value) -> DeckModelOut:
    """
    Search the database for the first deck that matches the given criteria.

    Args:
        property_name: The property to check.
        value: The value to match on.

    Returns:
        The first matching deck, or None if none were found.
    """

    deck = await db_core.get_document_by_property(DbCollection.DECKS, property_name, value)
    if deck is not None:
        return DeckModelOut.model_validate(deck)


async def update_deck(id: str, deck: DeckModelIn) -> DeckModelOut:
    """
    Update a deck in the database with the given values.

    Args:
        id: The ID of the deck to update.
        deck: The properties to update and their new values.

    Returns:
        The updated deck, or None if it was unable to update or find the deck.
    """

    updated_deck = await db_core.update_document(
        DbCollection.DECKS,
        id,
        deck.model_dump(
            mode="json",
            by_alias=True,
            include=deck.model_fields_set,
        ),
    )

    if updated_deck is not None:
        return DeckModelOut.model_validate(updated_deck)


async def delete_deck(id: str) -> DeckModelOut:
    """
    Delete a deck from the database.

    Args:
        id: The ID of the deck to delete.

    Returns:
        The deleted deck, or None if unable to delete or find the deck.
    """

    deleted_deck = await db_core.delete_document(DbCollection.DECKS, id)

    if deleted_deck is not None:
        return DeckModelOut.model_validate(deleted_deck)


# endregion

# region Decks


async def add_decks(decks: list[DeckModelIn]) -> list[ObjectId]:
    """
    Add a list of decks to the database.

    Args:
        decks: The list of deck to insert.

    Returns:
        The list of IDs for decks that were inserted, or empty list if unable.
    """

    insert_many_result = await db_core.insert_many_documents(
        DbCollection.DECKS,
        [
            deck.model_dump(
                mode="json",
                by_alias=True,
            )
            for deck in decks
        ],
    )

    return insert_many_result.inserted_ids if insert_many_result is not None else []


async def get_decks_random(limit: int) -> list[DeckModelOut]:
    """
    Get a random assortment of decks from the database.

    Args:
        limit: The number of decks to return.

    Returns:
        A random list of decks, or empty list if none were found.
    """

    decks = await db_core.get_random_documents(DbCollection.DECKS, limit)
    return [DeckModelOut.model_validate(deck) for deck in decks]


async def get_decks_by_property(
    property_name: str, values: list[Any], paginated: bool = False, page: int = 1, page_size: int = 10
) -> list[DeckModelOut]:
    """
    Search the database for decks matching the given criteria, with options for
    pagination.

    Args:
        property_name: The property to check.
        values: A list of values to match on.
        paginated: Whether to paginate the results.
        page: The page to look at, if paginated.
        page_size: The size of each page, if paginated.

    Returns:
        A list of decks matching the search criteria, or empty list if none
        were found.
    """

    decks = await db_core.get_documents_by_property(
        DbCollection.DECKS, property_name, values, paginated, page, page_size
    )

    return [DeckModelOut.model_validate(deck) for deck in decks]


async def delete_decks_all() -> int:
    """
    Delete all decks in the database.

    Returns:
        The number of decks deleted, or None if none could be deleted.
    """

    delete_many_result = await db_core.delete_documents(DbCollection.DECKS)
    if delete_many_result is not None:
        return delete_many_result.deleted_count


# endregion



import json
import re
from datetime import date
from typing import Iterable, Mapping, Self, TypeVar

from scooze.cardparts import (
    CardFace,
    CardPartsNormalizer,
    FullCardFace,
    ImageUris,
    Preview,
    Prices,
    PurchaseUris,
    RelatedCard,
    RelatedUris,
)
from scooze.catalogs import (
    BorderColor,
    Color,
    Finish,
    Format,
    Frame,
    FrameEffect,
    Game,
    ImageStatus,
    Language,
    Layout,
    Legality,
    Rarity,
    SecurityStamp,
    SetType,
)
from scooze.models.card import CardModelOut
from scooze.utils import FloatableT, HashableObject

## Generic Types
CardFaceT = TypeVar("CardFaceT", bound=CardFace)  # generic CardFace type


class Card(HashableObject):
    """
    A basic Card object with minimal fields. Contains all information you might
    use to sort a decklist.

    Attributes:
        scooze_id: A unique identifier for a document in a scooze database.
        cmc: This card's mana value/converted mana cost.
        color_identity: This card's color identity, for Commander variant
          deckbuilding.
        colors: This card's colors.
        legalities: Formats and the legality status of this card in them.
        mana_cost: Mana cost, as string of mana symbols.
          (e.g. "{1}{W}{U}{B}{R}{G}")
        name: This card's name.
        power: Power of this card, if applicable.
        toughness: Toughness of this card, if applicable.
        type_line: This card's type line. (e.g. "Creature — Ooze")
    """

    def __init__(
        self,
        cmc: FloatableT | None = None,
        color_identity: Iterable[Color] | None = None,
        colors: Iterable[Color] | None = None,
        legalities: Mapping[Format, Legality] | None = None,
        mana_cost: str | None = None,
        name: str | None = None,
        power: str | None = None,
        toughness: str | None = None,
        type_line: str | None = None,
        # kwargs
        **kwargs,  # TODO(77): log information about kwargs
    ):
        self.scooze_id = kwargs.get("scooze_id")

        self.cmc = CardNormalizer.to_float(cmc)
        self.color_identity = CardNormalizer.to_frozenset(color_identity, convert_to_enum=Color)
        self.colors = CardNormalizer.to_frozenset(colors, convert_to_enum=Color)
        self.legalities = CardNormalizer.to_frozendict(
            legalities, convert_key_to_enum=Format, convert_value_to_enum=Legality
        )
        self.mana_cost = mana_cost
        self.name = name
        self.power = power
        self.toughness = toughness
        self.type_line = type_line

    def __str__(self):
        return self.name

    @classmethod
    def from_json(cls, data: dict | str) -> Self:
        if isinstance(data, dict):
            return cls(**data)
        elif isinstance(data, str):
            return cls(**json.loads(data))

    @classmethod
    def from_model(cls, model: CardModelOut) -> Self:
        return cls(**model.model_dump())


class OracleCard(Card):
    """
    Card subclass containing all information about a unique card in Magic.
    All information in this class is print-agnostic.

    Attributes:
        scooze_id: A unique identifier for a document in a scooze database.
        card_faces: All component CardFace objects of this card, for multifaced
          cards.
        cmc: This card's mana value/converted mana cost.
        color_identity: This card's color identity, for Commander variant
          deckbuilding.
        color_indicator: The colors in this card's color indicator, if it has
          one.
        colors: This card's colors.
        edhrec_rank: This card's rank/popularity on EDHREC, if applicable.
        hand_modifier: This card's Vanguard hand size modifier, if applicable.
        keywords: Keywords and keyword actions this card uses.
        legalities: Formats and the legality status of this card in them.
        life_modifier: This card's Vanguard life modifier value, if applicable.
        loyalty: This card's starting planeswalker loyalty, if applicable.
        mana_cost: Mana cost, as string of mana symbols.
          (e.g. "{1}{W}{U}{B}{R}{G}")
        name: This card's name.
        oracle_id: A UUID for this card's oracle identity; shared across prints
          of the same card but not same-named objects with different gameplay
          properties.
        oracle_text: This card's oracle text, if any.
        penny_rank: This card's rank/popularity on Penny Dreadful.
        power: Power of this card, if applicable.
        prints_search_uri: A link to begin paginating through all prints of
          this card in Scryfall's API.
        produced_mana: Which colors of mana this card can produce.
        reserved: Whether this card is on the Reserved List.
        rulings_uri: A link to rulings for this card in Scryfall's API.
        toughness: Toughness of this card, if applicable.
        type_line: This card's type line. (e.g. "Creature — Ooze")
    """

    def __init__(
        self,
        card_faces: Iterable[CardFace] | None = None,
        cmc: FloatableT | None = None,
        color_identity: Iterable[Color] | None = None,
        color_indicator: Iterable[Color] | None = None,
        colors: Iterable[Color] | None = None,
        edhrec_rank: int | None = None,
        hand_modifier: str | None = None,
        keywords: Iterable[str] = None,
        legalities: Mapping[Format, Legality] = None,
        life_modifier: str | None = None,
        loyalty: str | None = None,
        mana_cost: str | None = None,
        name: str | None = None,
        oracle_id: str | None = None,
        oracle_text: str | None = None,
        penny_rank: int | None = None,
        power: str | None = None,
        prints_search_uri: str | None = None,
        produced_mana: Iterable[Color] | None = None,
        reserved: bool | None = None,
        rulings_uri: str | None = None,
        toughness: str | None = None,
        type_line: str | None = None,
        # kwargs
        **kwargs,  # TODO(77): log information about kwargs
    ):
        super().__init__(
            cmc=cmc,
            color_identity=color_identity,
            colors=colors,
            legalities=legalities,
            mana_cost=mana_cost,
            name=name,
            power=power,
            toughness=toughness,
            type_line=type_line,
            **kwargs,
        )
        self.scooze_id = kwargs.get("scooze_id")

        self.card_faces = CardNormalizer.to_card_faces(card_faces, card_face_class=CardFace)
        self.color_indicator = CardNormalizer.to_frozenset(color_indicator, convert_to_enum=Color)
        self.edhrec_rank = edhrec_rank
        self.hand_modifier = hand_modifier
        self.keywords = CardNormalizer.to_frozenset(keywords)
        self.life_modifier = life_modifier
        self.loyalty = loyalty
        self.oracle_id = oracle_id
        self.oracle_text = oracle_text
        self.penny_rank = penny_rank
        self.prints_search_uri = prints_search_uri
        self.produced_mana = CardNormalizer.to_frozenset(produced_mana, convert_to_enum=Color)
        self.reserved = reserved
        self.rulings_uri = rulings_uri

    @classmethod
    def oracle_text_without_reminder(cls, oracle_text: str) -> str:
        """
        Provide the given oracle text with reminder text removed.

        Args:
            oracle_text: The oracle text of a card.
        """

        pattern_reminder = r" ?\([^()]+\) ?"  # text between parens ()
        return re.sub(pattern_reminder, "", oracle_text)

    def is_double_sided(self) -> bool:
        """
        Determine if this is a double-sided card.
        """

        return self.card_faces is not None

    def total_words(self) -> int:
        """
        The number of words in this card's oracle text (excludes reminder text).
        """

        pattern_words = r"([a-zA-Z0-9+/{}']+)"  # words on a card

        # MDFC
        if self.is_double_sided():
            return sum(
                [
                    len(re.findall(pattern_words, OracleCard.oracle_text_without_reminder(face.oracle_text)))
                    for face in self.card_faces
                ]
            )
        # Non-MDFC
        else:
            return len(re.findall(pattern_words, OracleCard.oracle_text_without_reminder(self.oracle_text)))


class FullCard(OracleCard):
    """
    Card object that supports all fields available from Scryfall's JSON data.
    Scryfall documentation: https://scryfall.com/docs/api/cards

    Attributes:
    scooze_id: A unique identifier for a document in a scooze database.

    Core fields
        arena_id: This card's Arena ID, if applicable.
        scryfall_id: Scryfall's unique ID for this card.
        lang: The language code for this print;
          see https://scryfall.com/docs/api/languages
        mtgo_id: This card's MTGO Catalog ID, if applicable.
        mtgo_foil_id: This card's foil MTGO Catalog ID, if applicable.
        multiverse_ids: This card's multiverse IDs on Gatherer, if any.
        tcgplayer_id: This card's ID on TCGplayer, or `productId` in their
          system.
        tcgplayer_etched_id: This card's ID on TCGplayer, for the etched
          version if that is a separate product.
        cardmarket_id: This card's ID on Cardmarket, or `idProduct` in their
          system.
        oracle_id: A UUID for this card's oracle identity; shared across prints
          of the same card but not same-named objects with different gameplay
          properties.
        prints_search_uri: A link to begin paginating through all prints of
          this card in Scryfall's API.
        rulings_uri: A link to rulings for this card in Scryfall's API.
        scryfall_uri: A link to the Scryfall page for this card.
        uri: A link to this card object in Scryfall's API.

    Gameplay fields
        all_parts: RelatedCard objects for tokens/meld pairs/other associated
          parts to this card, if applicable.
        card_faces: All component CardFace objects of this card, for multifaced
          cards.
        cmc: This card's mana value/converted mana cost.
        color_identity: This card's color identity, for Commander variant
          deckbuilding.
        color_indicator: color_indicator: The colors in this card's color
          indicator, if it has one.
        colors: This card's colors.
        edhrec_rank: This card's rank/popularity on EDHREC, if applicable.
        hand_modifier: This card's Vanguard hand size modifier, if applicable.
        keywords: Keywords and keyword actions this card uses.
        legalities: Formats and the legality status of this card in them.
        life_modifier: This card's Vanguard life modifier value, if applicable.
        loyalty: This card's starting planeswalker loyalty, if applicable.
        mana_cost: Mana cost, as string of mana symbols.
          (e.g. "{1}{W}{U}{B}{R}{G}")
        name: This card's name.
        oracle_text: This card's oracle text, if any.
        penny_rank: This card's rank/popularity on Penny Dreadful.
        power: Power of this card, if applicable.
        produced_mana: Which colors of mana this card can produce.
        reserved: Whether this card is on the Reserved List.
        toughness: Toughness of this card, if applicable.
        type_line: This card's type line. (e.g. "Creature — Ooze")

    Print fields
        artist: Artist for this card.
        artist_ids: List of Scryfall IDs for artists of this card.
        attraction_lights: Attraction lights lit on this card, if applicable.
        booster: Whether this card can be opened in booster packs.
        border_color: Border color of this card, from among
          black, white, borderless, silver, and gold.
        card_back_id: Scryfall UUID of the card back design for this card.
        collector_number: This card's collector number; can contain non-numeric
          characters.
        content_warning: True if use of this print should be avoided;
          see https://scryfall.com/blog/regarding-wotc-s-recent-statement-on-depictions-of-racism-220
        digital: True if this card was only released in a video game.
        finishes: Finishes this card is available in, from among foil, nonfoil, and etched.
        flavor_name: Alternate name for this card, such as on Godzilla series.
        flavor_text: Flavor text on this card, if any.
        frame_effects: Special frame effects on this card;
          see https://scryfall.com/docs/api/frames
        frame: This card's frame layout;
          see https://scryfall.com/docs/api/frames
        full_art: Whether this print is full-art.
        games: Which games this print is available on, from among
          paper, mtgo, and arena.
        highres_image: Whether this card has a high-res image available.
        illustration_id: A UUID for the particular artwork on this print,
          consistent across art reprints.
        image_status: The quality/status of images available for this card.
          Either missing, placeholder, lowres, or highres_scan.
        image_uris: Links to images of this card in various qualities.
        layout: This card's printed layout;
          see https://scryfall.com/docs/api/layouts
        oversized: Whether this card is oversized.
        preview: Information about where, when, and how this print was
          previewed.
        prices: Prices for this card on various marketplaces.
        printed_name: Printed name of this card, for localized non-English
          cards.
        printed_text: Printed text of this card, for localized non-English
          cards.
        printed_type_line: Printed type line of this card, for localized
          non-English cards.
        promo: Whether this print is a promo.
        promo_types: Which promo categories this print falls into, if any.
        purchase_uris: Links to purchase this print from marketplaces.
        rarity: The rarity of this print.
        related_uris: Links to this print's listing on other online resources.
        released_at: The date this card was first released.
        reprint: Whether this print is a reprint from an earlier set.
        scryfall_set_uri: Link to the Scryfall set page for the set of this
          print.
        security_stamp: Security stamp on this card, if any.
        set_name: Full name of the set this print belongs to.
        set_search_uri: Link to Scryfall API to start paginating through this
          print's full set.
        set_type: An overall categorization for each set, provided by Scryfall.
        set_uri: Link to the set object for this print in Scryfall's API.
        set: Set code of the set this print belongs to.
        set_id: UUID of the set this print belongs to.
        story_spotlight: Whether this print is a Story Spotlight.
        textless: Whether this print is textless.
        variation: Whether this card print is a variation of another card
          object.
        variation_of: Which card object this object is a variant of, if any.
        watermark: Watermark printed on this card, if any.
    """

    def __init__(
        self,
        # Aliases
        id: str = "",  # Alias for scryfall_id
        # Core Fields
        arena_id: int | None = None,
        scryfall_id: str | None = None,
        lang: Language | None = None,
        mtgo_id: int | None = None,
        mtgo_foil_id: int | None = None,
        multiverse_ids: Iterable[int] | None = None,
        tcgplayer_id: int | None = None,
        tcgplayer_etched_id: int | None = None,
        cardmarket_id: int | None = None,
        oracle_id: str | None = None,
        prints_search_uri: str | None = None,
        rulings_uri: str | None = None,
        scryfall_uri: str | None = None,
        uri: str | None = None,
        # Gameplay Fields
        all_parts: Iterable[RelatedCard] | None = None,
        card_faces: Iterable[FullCardFace] | None = None,
        cmc: FloatableT | None = None,
        color_identity: Iterable[Color] | None = None,
        color_indicator: Iterable[Color] | None = None,
        colors: Iterable[Color] | None = None,
        edhrec_rank: int | None = None,
        hand_modifier: str | None = None,
        keywords: Iterable[str] | None = None,
        legalities: Mapping[Format, Legality] | None = None,
        life_modifier: str | None = None,
        loyalty: str | None = None,
        mana_cost: str | None = None,
        name: str | None = None,
        oracle_text: str | None = None,
        oversized: bool | None = None,
        penny_rank: int | None = None,
        power: str | None = None,
        produced_mana: Iterable[Color] | None = None,
        reserved: bool | None = None,
        toughness: str | None = None,
        type_line: str | None = None,
        # Print Fields
        artist: str | None = None,
        artist_ids: Iterable[str] | None = None,
        attraction_lights: Iterable[int] | None = None,
        booster: bool | None = None,
        border_color: BorderColor | None = None,
        card_back_id: str | None = None,
        collector_number: str | None = None,
        content_warning: bool = False,
        digital: bool | None = None,
        finishes: Iterable[Finish] | None = None,
        flavor_name: str | None = None,
        flavor_text: str | None = None,
        frame_effects: Iterable[FrameEffect] | None = None,
        frame: Frame | None = None,
        full_art: bool | None = None,
        games: Iterable[Game] | None = None,
        highres_image: bool | None = None,
        illustration_id: str | None = None,
        image_status: ImageStatus | None = None,
        image_uris: ImageUris | dict | None = None,
        layout: Layout | None = None,
        preview: Preview | None = None,
        prices: Prices | dict | None = None,
        printed_name: str | None = None,
        printed_text: str | None = None,
        printed_type_line: str | None = None,
        promo: bool | None = None,
        promo_types: Iterable[str] | None = None,
        purchase_uris: PurchaseUris | None = None,
        rarity: Rarity | None = None,
        related_uris: RelatedUris | None = None,
        released_at: date | str | None = None,
        reprint: bool | None = None,
        scryfall_set_uri: str | None = None,
        security_stamp: SecurityStamp | None = None,
        set_name: str | None = None,
        set_search_uri: str | None = None,
        set_type: SetType | None = None,
        set_uri: str | None = None,
        set: str | None = None,
        set_id: str | None = None,
        story_spotlight: bool | None = None,
        textless: bool | None = None,
        variation: bool | None = None,
        variation_of: str | None = None,
        watermark: str | None = None,
        # kwargs
        **kwargs,  # TODO(77): log information about kwargs
    ):
        super().__init__(
            card_faces=None,  # will be overridden with FullCardFace objects
            cmc=0,  # will be overridden with reversible card logic
            color_identity=color_identity,
            color_indicator=color_indicator,
            colors=colors,
            edhrec_rank=edhrec_rank,
            hand_modifier=hand_modifier,
            keywords=keywords,
            legalities=legalities,
            life_modifier=life_modifier,
            loyalty=loyalty,
            mana_cost=mana_cost,
            name=name,
            oracle_id=oracle_id,
            oracle_text=oracle_text,
            penny_rank=penny_rank,
            power=power,
            prints_search_uri=prints_search_uri,
            produced_mana=produced_mana,
            reserved=reserved,
            rulings_uri=rulings_uri,
            toughness=toughness,
            type_line=type_line,
            **kwargs,
        )
        self.scooze_id = kwargs.get("scooze_id")

        # region Core Fields

        self.arena_id = arena_id
        self.scryfall_id = scryfall_id if scryfall_id else id
        self.lang = CardNormalizer.to_enum(Language, lang)
        self.mtgo_id = mtgo_id
        self.mtgo_foil_id = mtgo_foil_id
        self.multiverse_ids = CardNormalizer.to_tuple(multiverse_ids)
        self.tcgplayer_id = tcgplayer_id
        self.tcgplayer_etched_id = tcgplayer_etched_id
        self.cardmarket_id = cardmarket_id
        self.scryfall_uri = scryfall_uri
        self.uri = uri

        # endregion

        # region Gameplay Fields

        self.all_parts = CardNormalizer.to_all_parts(all_parts)
        self.card_faces = CardNormalizer.to_card_faces(card_faces, card_face_class=FullCardFace)
        # Reversible cards currently have the same oracle card on both sides; will need to change this if this changes.
        if layout == Layout.REVERSIBLE_CARD:
            self.cmc = CardNormalizer.to_float(self.card_faces[0].cmc)
        else:
            self.cmc = CardNormalizer.to_float(cmc)
        self.oversized = oversized

        # endregion

        # region Print fields

        self.artist = artist
        self.artist_ids = CardNormalizer.to_tuple(artist_ids)
        self.attraction_lights = CardNormalizer.to_frozenset(attraction_lights)
        self.booster = booster
        self.border_color = CardNormalizer.to_enum(BorderColor, border_color)
        self.card_back_id = card_back_id
        self.collector_number = collector_number
        self.content_warning = content_warning
        self.digital = digital
        self.finishes = CardNormalizer.to_frozenset(finishes, convert_to_enum=Finish)
        self.flavor_name = flavor_name
        self.flavor_text = flavor_text
        self.frame_effects = CardNormalizer.to_frozenset(frame_effects, convert_to_enum=FrameEffect)
        self.frame = CardNormalizer.to_enum(Frame, frame)
        self.full_art = full_art
        self.games = CardNormalizer.to_frozenset(games, convert_to_enum=Game)
        self.highres_image = highres_image
        self.illustration_id = illustration_id
        self.image_status = CardNormalizer.to_enum(ImageStatus, image_status)
        self.image_uris = CardNormalizer.to_image_uris(image_uris)
        self.layout = CardNormalizer.to_enum(Layout, layout)
        self.preview = CardNormalizer.to_preview(preview)
        self.prices = CardNormalizer.to_prices(prices)
        self.printed_name = printed_name
        self.printed_text = printed_text
        self.printed_type_line = printed_type_line
        self.promo = promo
        self.promo_types = CardNormalizer.to_frozenset(promo_types)
        self.purchase_uris = CardNormalizer.to_purchase_uris(purchase_uris)
        self.rarity = CardNormalizer.to_enum(Rarity, rarity)
        self.related_uris = CardNormalizer.to_related_uris(related_uris)
        self.released_at = CardNormalizer.to_date(released_at)
        self.reprint = reprint
        self.scryfall_set_uri = scryfall_set_uri
        self.security_stamp = CardNormalizer.to_enum(SecurityStamp, security_stamp)
        self.set_name = set_name
        self.set_search_uri = set_search_uri
        self.set_type = CardNormalizer.to_enum(SetType, set_type)
        self.set_uri = set_uri
        self.set = set
        self.set_id = set_id
        self.story_spotlight = story_spotlight
        self.textless = textless
        self.variation = variation
        self.variation_of = variation_of
        self.watermark = watermark

        # endregion

    def total_words(self) -> int:
        """
        The number of words in this card's oracle text (excludes reminder text).
        """

        # Don't double count reversible card text
        return int(super().total_words() / (2 if self.layout is Layout.REVERSIBLE_CARD else 1))


class CardNormalizer(CardPartsNormalizer):
    """
    A simple class to use when normalizing non-serializable data from JSON.

    Usage:
        >>> card.prices = CardNormalizer.prices(prices_json)
    """

    @classmethod
    def to_all_parts(cls, all_parts: Iterable[RelatedCard] | Iterable[dict] | None) -> tuple[RelatedCard] | None:
        """
        Normalize all_parts from JSON.

        Args:
            all_parts: An Iterable[RelatedCard] or Iterable[JSON] to normalize.

        Returns:
            A tuple[RelatedCard].
        """

        if all_parts is None or all(isinstance(part, RelatedCard) for part in all_parts):
            return all_parts
        elif all(isinstance(part, dict) for part in all_parts):
            return tuple(RelatedCard(**part) for part in all_parts)

    @classmethod
    def to_card_faces(
        cls,
        card_faces: Iterable[CardFaceT] | Iterable[dict] | None,
        card_face_class: type[CardFaceT] = CardFace,
    ) -> tuple[CardFaceT]:
        """
        Normalize card_faces from JSON.

        Args:
            card_faces: A Iterable[F] or Iterable[JSON] to normalize. F is of type
              CardFace or FullCardFace.
            card_face_class: A CardFace class to create an instance of.
              (one of CardFace or FullCardFace)

        Returns:
            A tuple[F] where F is of type CardFace or FullCardFace.
        """

        if card_faces is None or all(isinstance(card_face, card_face_class) for card_face in card_faces):
            return card_faces
        elif all(isinstance(card_face, dict) for card_face in card_faces):
            return tuple(card_face_class.from_json(card_face) for card_face in card_faces)

    @classmethod
    def to_preview(cls, preview: Preview | dict | None) -> Preview:
        """
        Normalize preview from JSON.

        Args:
            preview: An instance of Preview or some JSON to normalize.

        Returns:
            An instance of Preview.
        """

        if preview is None or isinstance(preview, Preview):
            return preview
        elif isinstance(preview, dict):
            return Preview(**preview)

    @classmethod
    def to_prices(cls, prices: Prices | dict | None) -> Prices:
        """
        Normalize prices from JSON.

        Args:
            prices: An instance of Prices or some JSON to normalize.

        Returns:
            An instance of Prices.
        """

        if prices is None or isinstance(prices, Prices):
            return prices
        elif isinstance(prices, dict):
            return Prices(**prices)


CardT = TypeVar("CardT", bound=Card)  # generic Card type



from typing import Any

import scooze.database.card as db
from fastapi import APIRouter
from fastapi.responses import JSONResponse
from scooze.models.card import CardModelIn

router = APIRouter(
    prefix="/cards",
    tags=["cards"],
    responses={404: {"description": "Cards Not Found"}},
)


@router.get("/", summary="Get cards at random")
async def cards_root(limit: int = 3):
    """
    Get random cards up to the given limit.

    - **limit** - the maximum number of cards to get
    """

    cards = await db.get_cards_random(limit=limit)
    if cards is not None:
        return JSONResponse([card.model_dump(mode="json") for card in cards], status_code=200)
    else:
        return JSONResponse({"message": "No cards found in the database."}, status_code=404)


# Create


@router.post("/add", summary="Create new cards")
async def add_cards(cards: list[CardModelIn]):
    inserted_ids = await db.add_cards(cards=cards)
    if inserted_ids is not None:
        return JSONResponse({"message": f"Created {len(inserted_ids)} card(s)."}, status_code=200)
    else:
        return JSONResponse({"message": f"Failed to create any new cards."}, status_code=400)


# Read


@router.post("/by", summary="Get cards by property")
async def get_cards_by(
    property_name: str,
    values: list[Any],
    paginated: bool = False,
    page: int = 1,
    page_size: int = 10,
):
    """
    Get cards where the given property matches any of the given values.

    - **property_name** - the property to check against
    - **values** - matching values of the given property
    - **paginated** - return paginated results if True, return all matches if
    False
    - **page** - return matches from the given page
    - **page_size** - the number of results per page
    """

    cards = await db.get_cards_by_property(
        property_name=property_name, values=values, paginated=paginated, page=page, page_size=page_size
    )
    if cards is not None:
        return JSONResponse([card.model_dump(mode="json") for card in cards], status_code=200)
    else:
        return JSONResponse({"message": "Cards not found."}, status_code=404)


# Delete


@router.delete("/delete/all", summary="Delete all cards")
async def delete_cards_all():
    deleted_count = await db.delete_cards_all()

    if deleted_count is not None:
        return JSONResponse({"message": f"Deleted {deleted_count} card(s)."}, status_code=200)
    else:
        return JSONResponse({"message": "No cards deleted."}, status_code=404)



import scooze.database.deck as db
from scooze.card import CardT
from scooze.deck import Deck

# TODO(#145): need to support creating Deck[CardT] from model


def get_deck_by(property_name: str, value, card_class: CardT) -> Deck[CardT]:
    raise NotImplementedError()


def add_deck(deck: Deck):
    raise NotImplementedError()


def delete_deck(id: str):
    raise NotImplementedError()



import os

import requests
from scooze.catalogs import ScryfallBulkFile
from scooze.utils import DEFAULT_BULK_FILE_DIR

SCRYFALL_BULK_INFO_ENDPOINT = "https://api.scryfall.com/bulk-data"


def download_bulk_data_file(
    uri: str,
    bulk_file_type: ScryfallBulkFile | None = None,
    bulk_file_dir: str = DEFAULT_BULK_FILE_DIR,
) -> None:
    """
    Download a single bulk data file from Scryfall.

    Args:
        uri: Location of bulk data file (generally found from bulk info
          endpoint).
        bulk_file_type: Type of bulk file, used to set filename.
        bulk_file_dir: Directory to save bulk files. Defaults to `./data/bulk` if
          not specified.

    Raises:
        HTTPError: If request for bulk file not successful.
    """

    # TODO(#74): flag for check vs existing file; don't overwrite with same file or older version
    with requests.get(uri, stream=True) as r:
        r.raise_for_status()
        os.makedirs(bulk_file_dir, exist_ok=True)
        file_name = f"{bulk_file_dir}/{bulk_file_type}.json"
        with open(file_name, "wb") as f:
            for chunk in r.iter_content(chunk_size=None):
                f.write(chunk)


def download_bulk_data_file_by_type(
    bulk_file_type: ScryfallBulkFile | None = None,
    bulk_file_dir: str = DEFAULT_BULK_FILE_DIR,
) -> None:
    """
    Get a bulk data file from Scryfall, specified by file type (from among ScryfallBulkFile).

    Args:
        bulk_file_type: Type of bulk file, used to set filename.
        bulk_file_dir: Directory to save bulk files. Defaults to `./data/bulk` if
          not specified.

    Raises:
        HTTPError: If request for bulk file not successful.
    """

    # get URI from Scryfall bulk endpoint

    with requests.get(SCRYFALL_BULK_INFO_ENDPOINT) as bulk_metadata_request:
        bulk_metadata_request.raise_for_status()
        bulk_metadata = bulk_metadata_request.json()["data"]
    bulk_files = {t["type"]: t["download_uri"] for t in bulk_metadata}
    if bulk_file_type not in bulk_files:
        return
    download_bulk_data_file(bulk_files[bulk_file_type], bulk_file_type, bulk_file_dir)


def download_all_bulk_data_files(
    bulk_file_dir: str = DEFAULT_BULK_FILE_DIR,
) -> None:
    """
    Download all supported Scryfall bulk data files to local filesystem.

    Args:
        bulk_file_dir: Directory to save bulk files. Defaults to `./data/bulk` if
          not specified.

    Raises:
        HTTPError: If request for bulk file not successful.
    """

    with requests.get(SCRYFALL_BULK_INFO_ENDPOINT) as bulk_metadata_request:
        bulk_metadata_request.raise_for_status()
        bulk_metadata = bulk_metadata_request.json()["data"]
    bulk_files = {t["type"]: t["download_uri"] for t in bulk_metadata}

    for bulk_type in ScryfallBulkFile.list():
        bulk_filename = bulk_files[bulk_type]
        download_bulk_data_file(
            uri=bulk_filename,
            bulk_file_type=bulk_type,
            bulk_file_dir=bulk_file_dir,
        )



from functools import cache


def _check_for_safe_context(function):
    """
    Wrapper to ensure an instance method of ScoozeApi is called in a safe
    context.
    """

    def wrapper_safe_context(self, *args, **kwargs):
        if not self.safe_context:
            raise RuntimeError("ScoozeApi used outside of 'with' context")
        return function(self, *args, **kwargs)

    return wrapper_safe_context


@cache
def _safe_cache(function):
    """
    Wrapper for cache to ensure incoming arguments are hashable.
    """

    def _convert_list_args_to_tuple(*args, **kwargs):
        # TODO(#207): additional behavior to convert non-hashable arguments to hashable to allow cache
        args = [tuple(x) if type(x) == list else x for x in args]
        kwargs = {k: tuple(x) if type(x) == list else x for k, x in kwargs.items()}
        result = function(*args, **kwargs)
        result = tuple(result) if type(result) == list else result
        return result

    return function



import subprocess

from cleo.commands.command import Command


class TeardownDockerCommand(Command):
    name = "teardown docker"
    description = "Teardown MongoDB Docker container."

    def handle(self):
        self.line("Tearing down Docker container scooze-mongodb...")
        p = subprocess.run(
            "docker kill scooze-mongodb && docker rm scooze-mongodb && docker image rm mongo:latest",
            stdout=subprocess.DEVNULL,
            stderr=subprocess.STDOUT,
            shell=True,
        )
        self.line("Done.")



from enum import Enum, EnumMeta, StrEnum, auto

# region Enum Extensions


class CaseInsensitiveEnumMeta(EnumMeta):
    """
    An extension of the classic Python EnumMeta to support case-insensitive
    fields.
    """

    def __getitem__(self, item):
        if isinstance(item, str):
            item = item.upper()
        return super().__getitem__(item)


class ExtendedEnum(Enum, metaclass=CaseInsensitiveEnumMeta):
    """
    An extension of the classic Python Enum to support additional
    functionality.
    """

    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))


# endregion

# region Card Enums


class BorderColor(ExtendedEnum, StrEnum):
    """
    A color that borders of Magic cards can be.
    """

    BLACK = auto()
    WHITE = auto()
    BORDERLESS = auto()
    SILVER = auto()
    GOLD = auto()


class Color(ExtendedEnum, StrEnum):
    """
    A color that Magic cards can be.
    """

    WHITE = "W"
    BLUE = "U"
    BLACK = "B"
    RED = "R"
    GREEN = "G"

    # Not a color, but Scryfall uses this in `produced_mana` as a type that can be produced.
    COLORLESS = "C"


class Component(ExtendedEnum, StrEnum):
    """
    A type of related object, used in Scryfall to link a card
    to other cards or tokens referenced by it.
    See https://scryfall.com/docs/api/cards for documentation.
    """

    MELD_PART = auto()
    MELD_RESULT = auto()
    TOKEN = auto()
    COMBO_PIECE = auto()  # Generally a card referenced directly in rules text, in either direction.


class Finish(ExtendedEnum, StrEnum):
    """
    A finish type that a Magic card can be printed with.
    """

    FOIL = auto()
    NONFOIL = auto()
    ETCHED = auto()


class Format(ExtendedEnum, StrEnum):
    """
    A Magic: the Gathering competitive format.
    """

    ALCHEMY = auto()
    BRAWL = auto()
    COMMANDER = auto()
    DUEL = auto()
    EXPLORER = auto()
    FUTURE = auto()
    GLADIATOR = auto()
    HISTORIC = auto()
    HISTORICBRAWL = auto()
    LEGACY = auto()
    MODERN = auto()
    OATHBREAKER = auto()
    OLDSCHOOL = auto()
    PAUPER = auto()
    PAUPERCOMMANDER = auto()
    PENNY = auto()
    PIONEER = auto()
    PREDH = auto()
    PREMODERN = auto()
    STANDARD = auto()
    VINTAGE = auto()
    # non-Scryfall formats
    LIMITED = auto()
    NONE = auto()


class Frame(ExtendedEnum, StrEnum):
    """
    A frame style for a Magic card, corresponding to the year of its design.
    """

    _1993 = "1993"
    _1997 = "1997"
    _2003 = "2003"
    _2015 = "2015"
    FUTURE = auto()


class FrameEffect(ExtendedEnum, StrEnum):
    """
    A frame effect on a Magic card that's different from the usual even border.
    """

    COLORSHIFTED = auto()
    COMPANION = auto()
    DEVOID = auto()
    DRAFT = auto()
    ETCHED = auto()
    EXTENDEDART = auto()
    FULLART = auto()
    INVERTED = auto()
    LEGENDARY = auto()
    LESSON = auto()
    MIRACLE = auto()
    NYXTOUCHED = auto()
    SHATTEREDGLASS = auto()
    SHOWCASE = auto()
    SNOW = auto()
    TEXTLESS = auto()
    TOMBSTONE = auto()

    # Double-faced card marks
    COMPASSLANDDFC = auto()
    CONVERTDFC = auto()
    FANDFC = auto()
    MOONELDRAZIDFC = auto()
    ORIGINPWDFC = auto()
    SUNMOONDFC = auto()
    UPSIDEDOWNDFC = auto()
    WAXINGANDWANINGMOONDFC = auto()


class Game(ExtendedEnum, StrEnum):
    """
    An official Magic game a card print can be available in.
    """

    PAPER = auto()
    ARENA = auto()
    MTGO = auto()
    ASTRAL = auto()
    SEGA = auto()


class ImageStatus(ExtendedEnum, StrEnum):
    """
    An indicator for whether a card's image exists on Scryfall,
    and how high quality the sourced image is.
    See https://scryfall.com/docs/api/images for documentation.
    """

    MISSING = auto()
    PLACEHOLDER = auto()
    LOWRES = auto()
    HIGHRES_SCAN = auto()


class Language(ExtendedEnum, StrEnum):
    """
    A language that Magic cards have been printed in, using Scryfall's language codes.
    See https://scryfall.com/docs/api/languages for documentation.
    """

    ENGLISH = "en"
    SPANISH = "es"
    FRENCH = "fr"
    GERMAN = "de"
    ITALIAN = "it"
    PORTUGUESE = "pt"
    JAPANESE = "ja"
    KOREAN = "ko"
    RUSSIAN = "ru"
    CHINESE_SIMPLIFIED = "zhs"
    CHINESE_TRADITIONAL = "zht"
    PHYREXIAN = "ph"

    # each used for a single promotional card
    ANCIENT_GREEK = "grc"
    ARABIC = "ar"
    HEBREW = "he"
    LATIN = "la"
    SANSKRIT = "sa"


class Layout(ExtendedEnum, StrEnum):
    """
    A layout that a Magic card can be printed with.
    """

    NORMAL = auto()
    ADVENTURE = auto()
    ART_SERIES = auto()
    AUGMENT = auto()
    BATTLE = auto()
    CLASS = auto()
    DOUBLE_FACED_TOKEN = auto()
    EMBLEM = auto()
    FLIP = auto()
    HOST = auto()
    LEVELER = auto()
    MELD = auto()
    MODAL_DFC = auto()
    MUTATE = auto()
    PLANAR = auto()
    PROTOTYPE = auto()
    REVERSIBLE_CARD = auto()
    SAGA = auto()
    SCHEME = auto()
    SPLIT = auto()
    TOKEN = auto()
    TRANSFORM = auto()
    VANGUARD = auto()


class Legality(ExtendedEnum, StrEnum):
    """
    String enum of different legalities that a card can have in a format.

    - legal: eligible to be played in a format, and not banned
    - not_legal: not eligible to be legal (never printed in right set/rarity)
    - banned: eligible to be legal, but specifically banned
    - restricted: only 1 copy allowed in a deck (specific to Vintage)
    """

    LEGAL = auto()
    NOT_LEGAL = auto()
    BANNED = auto()
    RESTRICTED = auto()


class Rarity(ExtendedEnum, StrEnum):
    """
    A rarity that a print of a Magic card can be.
    """

    COMMON = auto()
    UNCOMMON = auto()
    RARE = auto()
    SPECIAL = auto()
    MYTHIC = auto()
    BONUS = auto()


class ScryfallBulkFile(ExtendedEnum, StrEnum):
    """
    Canonical names for Scryfall bulk files.
    """

    ORACLE = "oracle_cards"
    ARTWORK = "unique_artwork"
    DEFAULT = "default_cards"
    ALL = "all_cards"
    # TODO(#26): support for Rulings file


class SecurityStamp(ExtendedEnum, StrEnum):
    """
    A holographic security stamp printed on the bottom of some
    Magic cards, generally rares and mythics.
    """

    OVAL = auto()  # default style
    ACORN = auto()  # eternal-legal cards in Un-sets
    ARENA = auto()
    CIRCLE = auto()  # Signature Spellbook style
    HEART = auto()  # My Little Pony
    TRIANGLE = auto()  # Universes Beyond


class SetType(ExtendedEnum, StrEnum):
    """
    A Scryfall-provided categorization for a set.
    See https://scryfall.com/docs/api/sets for documentation.
    """

    ALCHEMY = auto()
    ARCHENEMY = auto()
    ARSENAL = auto()
    BOX = auto()
    COMMANDER = auto()
    CORE = auto()
    DRAFT_INNOVATION = auto()
    DUEL_DECK = auto()
    EXPANSION = auto()
    FROM_THE_VAULT = auto()
    FUNNY = auto()
    MASTERPIECE = auto()
    MASTERS = auto()
    MEMORABILIA = auto()
    MINIGAME = auto()
    PLANECHASE = auto()
    PREMIUM_DECK = auto()
    PROMO = auto()
    SPELLBOOK = auto()
    STARTER = auto()
    TOKEN = auto()
    TREASURE_CHEST = auto()
    VANGUARD = auto()


# endregion

# region Deck Enums


class InThe(ExtendedEnum, StrEnum):
    """
    The location of a Card in a Deck.
    """

    MAIN = auto()
    SIDE = auto()
    CMDR = auto()


class DecklistFormatter(ExtendedEnum, StrEnum):
    """
    A method of formatting a decklist for external systems.
    """

    ARENA = auto()
    MTGO = auto()


# endregion


# region Database Enums


class DbCollection(ExtendedEnum, StrEnum):
    """
    Collections in the Scooze database.
    """

    CARDS = "cards"
    DECKS = "decks"


# endregion



from __future__ import annotations

from importlib import import_module
from typing import TYPE_CHECKING

import pkg_resources
from cleo.application import Application
from cleo.loaders.factory_command_loader import FactoryCommandLoader

if TYPE_CHECKING:
    from collections.abc import Callable

    from cleo.commands.command import Command

# Accepted scooze CLI commands
COMMANDS = [
    "delete",
    "run",
    # Load commands
    "load cards",
    "load decks",
    # Setup commands
    "setup docker",
    "setup local",
    # Teardown commands
    "teardown docker",
    "teardown local",
]


def load_scooze_command(name: str) -> Callable[[], Command]:
    def _scooze_command_factory() -> Command:
        cli_words = name.split(" ")
        module = import_module("scooze.console.commands." + ".".join(cli_words))
        scooze_command_class = getattr(module, "".join(c.title() for c in cli_words) + "Command")
        command: Command = scooze_command_class()
        return command

    return _scooze_command_factory


class ScoozeApplication(Application):
    def __init__(self):
        pkg_name = "scooze"
        super().__init__(pkg_name, f"{pkg_resources.get_distribution(pkg_name).version}")

        command_loader = FactoryCommandLoader({name: load_scooze_command(name) for name in COMMANDS})
        self.set_command_loader(command_loader)


def run_cli() -> int:
    exit_code: int = ScoozeApplication().run()
    return exit_code


if __name__ == "__main__":
    run_cli()



from collections import Counter
from sys import maxsize
from typing import Generic, Self

import scooze.utils as utils
from scooze.card import CardT
from scooze.catalogs import DecklistFormatter, Format, InThe, Legality
from scooze.deckpart import DeckDiff, DeckPart


class Deck(utils.ComparableObject, Generic[CardT]):
    """
    A class to represent a deck of Magic: the Gathering cards.

    Attributes:
        archetype: The archetype of this Deck.
        format: The format legality of the cards in this Deck.
        main: The main deck. Typically 60 cards minimum.
        side: The sideboard. Typically 15 cards maximum.
        cmdr: The command zone. Typically 1 or 2 cards in Commander formats.
    """

    def __init__(
        self,
        archetype: str | None = None,
        format: Format = Format.NONE,
        main: DeckPart[CardT] = None,
        side: DeckPart[CardT] = None,
        cmdr: DeckPart[CardT] = None,
    ):
        self.archetype = archetype
        self.format = format

        self.main = main if main is not None else DeckPart()
        self.side = side if side is not None else DeckPart()
        self.cmdr = cmdr if cmdr is not None else DeckPart()

    @property
    def cards(self) -> Counter[CardT]:
        return self.main.cards + self.side.cards + self.cmdr.cards

    def __str__(self):
        decklist = self.export()
        return f"""Archetype: {self.archetype}\n""" f"""Format: {self.format}\n""" f"""Decklist:\n{decklist}\n"""

    def average_cmc(self) -> float:
        """
        The average mana value of cards in this Deck.
        """

        # TODO(#112): Add type filters.
        # TODO(#113): Reversible cards do not have a top-level cmc. Assign one?

        total_cards = self.total_cards()

        if total_cards > 0:
            return self.total_cmc() / self.total_cards()
        return 0

    def average_words(self) -> float:
        """
        The average number of words across all oracle text on all cards in this
        Deck (excludes reminder text).
        """

        # TODO(#112): Add type filters.

        total_cards = self.total_cards()

        if total_cards > 0:
            return self.total_words() / self.total_cards()
        return 0

    def diff(self, other: Self) -> DeckDiff:
        """
        Generate a diff between this Deck and another.

        Args:
            other: The other Deck.

        Returns:
            A DeckDiff with keys for each deck part. Each contains a dict of
            each card in both decks and their counts.
        """

        return DeckDiff(
            main=self.main.diff(other.main),
            side=self.side.diff(other.side),
            cmdr=self.cmdr.diff(other.cmdr),
        )

    def decklist_equals(self, other: Self) -> bool:
        """
        Determine if this Deck contains exactly the same cards as another.

        Args:
            other: The other Deck.

        Returns:
            True if this Deck contains exactly the same cards as another, else
            False.
        """

        if self.total_cards() != other.total_cards():
            return False

        diff = self.diff(other)
        same_main = not bool(diff.main)
        same_side = not bool(diff.side)
        same_cmdr = not bool(diff.cmdr)

        return same_main and same_side and same_cmdr

    def export(self, export_format: DecklistFormatter = None) -> str:
        """
        Export this Deck as a string with the given DecklistFormatter.

        Args:
            export_format: The format of the exported Deck.

        Returns:
            A string containing the names and quantities of the cards in this
            Deck.
        """

        match export_format:
            case DecklistFormatter.ARENA:
                sb_prefix = "Sideboard\n"
                cmdr_prefix = "Commander\n"
                # TODO(#50): filter out cards that are not on Arena. Log a WARNING with those cards.
            case DecklistFormatter.MTGO:
                sb_prefix = "SIDEBOARD:\n"
                cmdr_prefix = ""
                # TODO(#50): filter out cards that are not on MTGO. Log a WARNING with those cards.
            case _:
                sb_prefix = "Sideboard\n"  # Default
                cmdr_prefix = "Commander\n"  # Default
        sb_prefix = "\n" + sb_prefix
        cmdr_suffix = "\n"

        # Build the decklist string
        main = str(self.main) if len(self.main) > 0 else ""
        side = (sb_prefix + str(self.side)) if len(self.side) > 0 else ""
        cmdr = (cmdr_prefix + str(self.cmdr) + cmdr_suffix) if len(self.cmdr) > 0 else ""
        decklist = f"{cmdr}{main}{side}"

        return decklist

    def is_legal(self, format: Format = None) -> bool:
        """
        Determine if this Deck is legal in the given format.

        Default checks against `self.Format`. If `self.Format` is unset, checks against `Format.NONE`.

        - For cards with `Legality.RESTRICTED`, only 1 or fewer may be present
        throughout all deck parts.
        - For cards with `Legality.LEGAL`, only N or fewer may be present
        throughout all deck parts where N is determined by the max quantity of
        a single cards allowed by the given format.
        - For cards with `Legality.BANNED` or `Legality.NOT_LEGAL`, none may be
        present throughout all deck parts.

        Args:
            format: The format to check against.
        """

        # Default
        if format is None:
            format = self.format if self.format is not None else Format.NONE

        # Check deck meets minimum size requirements
        if self.main.total() < utils.main_size(format)[0]:
            return False
        if self.side.total() < utils.side_size(format)[0]:
            return False
        if self.cmdr.total() < utils.cmdr_size(format)[0]:
            return False

        # Check card quantities do not exceed acceptable maximums
        for c, q in self.cards.items():
            c_legal = c.legalities[format] if format not in [Format.LIMITED, Format.NONE] else Legality.LEGAL

            if (c_legal is Legality.RESTRICTED and q > 1) or c_legal in [Legality.BANNED, Legality.NOT_LEGAL]:
                return False

            if q > utils.max_card_quantity(format) and q > utils.max_relentless_quantity(c.name):
                return False

        return True

    def total_cards(self) -> int:
        """
        The number of cards in this Deck.
        """

        return self.main.total() + self.side.total() + self.cmdr.total()

    def total_cmc(self) -> float:
        """
        The total mana value of cards in this Deck.
        """

        # TODO(#113): Reversible cards do not have a top-level cmc. Assign one?

        return sum([c.cmc * q for c, q in self.cards.items()])

    def total_words(self) -> int:
        """
        The number of words across all oracle text on all cards in this Deck
        (excludes reminder text).
        """

        return sum([c.total_words() * q for c, q in self.cards.items()])

    # region Mutating Methods

    def add_card(self, card: CardT, quantity: int = 1, in_the: InThe = InThe.MAIN) -> None:
        """
        Add a given quantity of a given card to this Deck.

        Args:
            card: The card to add.
            quantity: The number of copies of the card to be added.
            in_the: Where to add the card (main, side, etc.)
        """

        match in_the:
            case InThe.MAIN:
                self.main.add_card(card=card, quantity=quantity)
            case InThe.SIDE:
                self.side.add_card(card=card, quantity=quantity)
            case InThe.CMDR:
                self.cmdr.add_card(card=card, quantity=quantity)
            case _:
                pass  # TODO(#75): 'in' must be one of InThe.list()

    def add_cards(self, cards: Counter[CardT], in_the: InThe = InThe.MAIN) -> None:
        """
        Add the given cards to this Deck.

        Args:
            cards: The cards to add.
            in_the: Where to add the cards (main, side, etc.)
        """

        match in_the:
            case InThe.MAIN:
                self.main.add_cards(cards)
            case InThe.SIDE:
                self.side.add_cards(cards)
            case InThe.CMDR:
                self.cmdr.add_cards(cards)

    def remove_card(self, card: CardT, quantity: int = maxsize, in_the: InThe = InThe.MAIN) -> None:
        """
        Remove a given quantity of a given card from this Deck. If quantity is
        not provided, removes all copies.

        Args:
            card: The card to remove.
            quantity: The number of copies of the card to be removed.
            in_the: Where to remove the cards from (main, side, etc.)
        """

        # using counterA - counterB results in a new Counter with only positive results
        match in_the:
            case InThe.MAIN:
                self.main.remove_card(card=card, quantity=quantity)
            case InThe.SIDE:
                self.side.remove_card(card=card, quantity=quantity)
            case InThe.CMDR:
                self.cmdr.remove_card(card=card, quantity=quantity)
            case _:
                pass  # TODO(#75): failed to remove card

    def remove_cards(self, cards: Counter[CardT], in_the: InThe = InThe.MAIN) -> None:
        """
        Remove the given cards from this Deck.

        Args:
            cards: The cards to remove.
            in_the: Where to remove the cards from (main, side, etc.)
        """

        # using counterA - counterB results in a new Counter with only positive results
        match in_the:
            case InThe.MAIN:
                self.main.remove_cards(cards=cards)
            case InThe.SIDE:
                self.side.remove_cards(cards=cards)
            case InThe.CMDR:
                self.cmdr.remove_cards(cards=cards)
            case _:
                pass  # TODO(#75): failed to remove cards

    # endregion



from datetime import date

from pydantic import Field, field_validator
from scooze.cardparts import ImageUris
from scooze.catalogs import Color, Component, Layout
from scooze.models.utils import ScoozeBaseModel


class ImageUrisModel(ScoozeBaseModel):
    """
    URIs of images associated with this model on Scryfall.
    Scryfall documentation: https://scryfall.com/docs/api/images

    Attributes:
        png: Full card, high quality image with transparent background and
          rounded corners.
        border_crop: Full card image with corners and majority of border
          cropped out.
        art_crop: Rectangular crop to just art box; may not be perfect for
          cards with strange layouts.
        large: Large JPG image (672x936)
        normal: Medium JPG image (488x860)
        small: Small JPG image (146x204)
    """

    png: str | None = Field(
        default=None,
        description="Full card, high quality image with transparent background and rounded corners.",
    )
    border_crop: str | None = Field(
        default=None,
        description="Full card image with corners and majority of border cropped out.",
    )
    art_crop: str | None = Field(
        default=None,
        description="Rectangular crop to just art box; may not be perfect for cards with strange layouts.",
    )
    large: str | None = Field(
        default=None,
        description="Large JPG image (672x936)",
    )
    normal: str | None = Field(
        default=None,
        description="Medium JPG image (488x860)",
    )
    small: str | None = Field(
        default=None,
        description="Small JPG image (146x204)",
    )


class CardFaceModel(ScoozeBaseModel):
    """
    Model for a single face of a multi-faced CardModel.
    Multi-faced cards include MDFCs, split cards, aftermath, etc.

    Scryfall documentation: https://scryfall.com/docs/api/cards#card-face-objects

    Attributes:
        artist: Illustrator for art on this face.
        artist_id: Scryfall ID for the artist of this face.
        cmc: Mana value of this face.
        color_indicator: Color indicator on this face, if any.
        colors: Colors of this face.
        flavor_text: Flavor text of this face, if any.
        illustration_id: Scryfall illustration ID of this face, if any.
        image_uris: Scryfall illustration ID of this face, if any.
        layout: Layout of this face, if any.
        loyalty: Starting planeswalker loyalty of this face, if any.
        mana_cost: Mana cost of this face.
        name: Name of this face.
        oracle_id: Oracle ID of this face, for reversible cards.
        oracle_text: Oracle text of this face, if any.
        power: Power of this face, if any.
        printed_name: Printed name of this face, for localized non-English
          cards.
        printed_text: Printed text of this face, for localized non-English
          cards.
        printed_type_line: Printed type line of this face, for localized
          non-English cards.
        toughness: Toughness of this face, if any.
        type_line: Type line of this face, if any.
        watermark: Watermark printed on this face, if any.
    """

    artist: str | None = Field(
        default=None,
        description="Illustrator for art on this face.",
    )
    artist_id: str | None = Field(
        default=None,
        description="Scryfall ID for the artist of this face.",
    )
    cmc: float | None = Field(
        default=None,
        description="Mana value of this face.",
    )
    color_indicator: set[Color] | None = Field(
        default=None,
        description="Color indicator on this face, if any.",
    )
    colors: set[Color] | None = Field(
        default=None,
        description="Colors of this face.",
    )
    flavor_text: str | None = Field(
        default=None,
        description="Flavor text of this face, if any.",
    )
    illustration_id: str | None = Field(
        default=None,
        description="Scryfall illustration ID of this face, if any.",
    )
    image_uris: ImageUrisModel | None = Field(
        default=None,
        description="URIs for images of this face on Scryfall.",
    )
    layout: Layout | None = Field(
        default=None,
        description="Layout of this face, if any.",
    )
    loyalty: str | None = Field(
        default=None,
        description="Starting planeswalker loyalty of this face, if any.",
    )
    mana_cost: str = Field(
        default="",
        description="Mana cost of this face.",
    )
    name: str = Field(
        description="Name of this face.",
    )
    oracle_id: str | None = Field(
        default=None,
        description="Oracle ID of this face, for reversible cards.",
    )
    oracle_text: str | None = Field(
        default=None,
        description="Oracle text of this face, if any.",
    )
    power: str | None = Field(
        default=None,
        description="Power of this face, if any.",
    )
    printed_name: str | None = Field(
        default=None,
        description="Printed name of this face, for localized non-English cards.",
    )
    printed_text: str | None = Field(
        default=None,
        description="Printed text of this face, for localized non-English cards.",
    )
    printed_type_line: str | None = Field(
        default=None,
        description="Printed type line of this face, for localized non-English cards.",
    )
    toughness: str | None = Field(
        default=None,
        description="Toughness of this face, if any.",
    )
    type_line: str | None = Field(
        default=None,
        description="Type line of this face, if any.",
    )
    watermark: str | None = Field(
        default=None,
        description="Watermark printed on this face, if any.",
    )

    # region Validators

    @field_validator("image_uris", mode="before")
    @classmethod
    def image_uris_validator(cls, v):
        if isinstance(v, ImageUris):
            v = ImageUrisModel.model_validate(v.__dict__)
        return v

    # endregion


class PricesModel(ScoozeBaseModel):
    """
    Model for all price data associated with a CardModel.

    Attributes:
        usd: Price in US dollars, from TCGplayer.
        usd_foil: Foil price in US dollars, from TCGplayer.
        usd_etched: Etched foil price in US dollars, from TCGplayer.
        eur: Price in Euros, from Cardmarket.
        eur_foil: Foil price in Euros, from Cardmarket.
        tix: Price in MTGO tix, from Cardhoarder.
    """

    usd: float | None = Field(
        default=None,
        description="Price in US dollars, from TCGplayer.",
    )
    usd_foil: float | None = Field(
        default=None,
        description="Foil price in US dollars, from TCGplayer.",
    )
    usd_etched: float | None = Field(
        default=None,
        description="Etched foil price in US dollars, from TCGplayer.",
    )
    eur: float | None = Field(
        default=None,
        description="Price in Euros, from Cardmarket.",
    )
    eur_foil: float | None = Field(
        default=None,
        description="Foil price in Euros, from Cardmarket.",
    )
    tix: float | None = Field(
        default=None,
        description="Price in MTGO tix, from Cardhoarder.",
    )


class PreviewModel(ScoozeBaseModel):
    """
    Object for information about where and when a card was previewed.

    Attributes:
        previewed_at: Date/time of preview being shown or added to Scryfall.
        source: Name of preview source.
        source_uri: Location of preview source.
    """

    previewed_at: date | None = Field(
        default=None,
        description="Date/time of preview being shown or added to Scryfall.",
    )
    source: str | None = Field(
        default=None,
        description="Name of preview source.",
    )
    source_uri: str | None = Field(
        default=None,
        description="Location of preview source.",
    )


class PurchaseUrisModel(ScoozeBaseModel):
    """
    URIs to this card’s listing on major marketplaces.

    Attributes:
        tcgplayer: Link to buy this card on the TCGplayer marketplace.
        cardmarket: Link to buy this card on the Cardmarket marketplace.
        cardhoarder: Link to buy this card digitally for MTGO on Cardhoarder.
    """

    tcgplayer: str | None = Field(
        default=None,
        description="Link to buy this card on the TCGplayer marketplace.",
    )
    cardmarket: str | None = Field(
        default=None,
        description="Link to buy this card on the Cardmarket marketplace.",
    )
    cardhoarder: str | None = Field(
        default=None,
        description="Link to buy this card digitally for MTGO on Cardhoarder.",
    )


class RelatedCardModel(ScoozeBaseModel):
    """
    Data about Scryfall objects related to this card
    (tokens, cards referenced by name, meld pairs, etc.)

    Scryfall documentation: https://scryfall.com/docs/api/cards#related-card-objects

    Attributes:
        scryfall_id: ID of linked component.
        component: One of `token`, `meld_part`, `meld_result`, or
          `combo_piece`.
        name: Name of linked component.
        type_line: Type line of linked component.
        uri: URI of linked component.
    """

    scryfall_id: str = Field(
        description="ID of linked component.",
        alias="id",
    )  # Scryfall ID
    component: Component = Field(
        description="One of `token`, `meld_part`, `meld_result`, or `combo_piece`.",
    )
    name: str = Field(
        description="Name of linked component.",
    )
    type_line: str = Field(
        description="Type line of linked component.",
    )
    uri: str = Field(
        description="URI of linked component.",
    )


class RelatedUrisModel(ScoozeBaseModel):
    """
    Links to information about a Scryfall-based card object on other non-Scryfall resources.

    Attributes:
        edhrec: Information about this card on edhrec.
        gatherer: Information about this card on Gatherer.
        tcgplayer_infinite_articles: Articles about this card on TCGplayer Infinite.
        tcgplayer_infinite_decks: Decks including this card on TCGplayer Infinite.
    """

    edhrec: str | None = Field(
        default=None,
        description="Information about this card on edhrec.",
    )
    gatherer: str | None = Field(
        default=None,
        description="Information about this card on Gatherer.",
    )
    tcgplayer_infinite_articles: str | None = Field(
        default=None,
        description="Articles about this card on TCGplayer Infinite.",
    )
    tcgplayer_infinite_decks: str | None = Field(
        default=None,
        description="Decks including this card on TCGplayer Infinite.",
    )



from typing import Any

from bson import ObjectId
from pymongo import ReturnDocument
from pymongo.results import DeleteResult
from scooze.catalogs import DbCollection
from scooze.database.mongo import db
from scooze.models.utils import _to_lower_camel

# region Single document


async def insert_document(coll_type: DbCollection, document: dict[str, Any]):
    """
    Insert a document into a collection in the database.

    Args:
        coll_type: The collection to insert into.
        document: The document to insert.

    Returns:
        The inserted document, or None if unable to insert.
    """

    # Here we find and update with upsert=True instead of inserting to avoid creating duplicates in the database. This
    # creates fewer headaches if you doubleclick in the Swagger UI or similar.
    return await db.client.scooze[coll_type].find_one_and_update(
        document,
        {"$setOnInsert": document},
        upsert=True,
        return_document=ReturnDocument.AFTER,
    )


async def get_document_by_property(coll_type: DbCollection, property_name: str, value):
    """
    Search a collection in the database for the first document matching the given criteria.

    Args:
        coll_type: The collection to search.
        property_name: The property to check.
        value: The value to match on.

    Returns:
        The first matching document, or None if none were found.
    """

    # Alias scooze_id to _id
    match property_name:
        case "_id" | "scooze_id":
            property_name = "_id"
            value = ObjectId(value)
        case _:
            property_name = _to_lower_camel(property_name)

    return await db.client.scooze[coll_type].find_one({property_name: value})


async def update_document(coll_type: DbCollection, id: str, document: dict[str, Any]):
    """
    Update a document in a collection with new values.

    Args:
        coll_type: The collection containing the document to update.
        id: The ID of the document to update.
        document: The properties to update and their new values.

    Returns:
        The updated document, or None if it was unable to update or find it.

    Raises:
        ValueError: No data was given to update.
    """

    if len(document) == 0:
        raise ValueError(f"No data given, skipping update for {coll_type.title} with id: {id}")
    return await db.client.scooze[coll_type].find_one_and_update(
        {"_id": ObjectId(id)},
        {"$set": document},
        return_document=ReturnDocument.AFTER,
    )


async def delete_document(coll_type: DbCollection, id: str):
    """
    Delete a document from the database.

    Args:
        coll_type: The collection to delete from.
        id: The ID of the document to delete.

    Returns:
        The deleted document, or None if unable to delete.
    """

    return await db.client.scooze[coll_type].find_one_and_delete({"_id": ObjectId(id)})


# endregion

# region Bulk operations


async def insert_many_documents(coll_type: DbCollection, documents: list[dict[str, Any]]):
    """
    Insert a list of documents into the database.

    Args:
        coll_type: The collection to insert into.
        documents: The list of documents to insert.

    Returns:
        A PyMongo InsertManyResult.
    """

    return await db.client.scooze[coll_type].insert_many(documents)


async def get_random_documents(coll_type: DbCollection, limit: int):
    """
    Get a random sample of documents from a single collection in the database.

    Args:
        coll_type: The desired collection.
        limit: The number of documents to return.

    Returns:
        The list of random documents.
    """

    pipeline = [{"$sample": {"size": limit}}]
    return await db.client.scooze[coll_type].aggregate(pipeline).to_list(limit)


async def get_all_documents(coll_type: DbCollection):
    """
    Get a random sample of documents from a single collection in the database.

    Args:
        coll_type: The desired collection.

    Returns:
        The list of random documents.
    """

    return await db.client.scooze[coll_type].find().to_list(None)


async def get_documents_by_property(
    coll_type: DbCollection,
    property_name: str,
    values: list[Any],
    paginated: bool = False,
    page: int = 1,
    page_size: int = 10,
):
    """
    Search the database for documents matching the given criteria, with options for pagination.

    Args:
        coll_type: The collection to read from.
        property_name: The property to check.
        values: A list of values to match on.
        paginated: Whether to paginate the results.
        page: The page to look at, if paginated.
        page_size: The size of each page, if paginated.

    Returns:
        A list of matching documents, or None if none were found.
    """

    # Alias scooze_id to _id
    match property_name:
        case "_id" | "scooze_id":
            property_name = "_id"
            vals = [ObjectId(i) for i in values]  # Handle ObjectIds
        case _:
            property_name = _to_lower_camel(property_name)
            vals = values

    return (
        await db.client.scooze[coll_type]
        .find({"$or": [{property_name: v} for v in vals]})
        .skip((page - 1) * page_size if paginated else 0)
        .to_list(page_size if paginated else None)
    )


async def delete_documents_by_id(coll_type: DbCollection, ids: list[ObjectId]) -> DeleteResult:
    """
    Deletes multiple documents from the database with the given IDs.

    Args:
        coll_type: The collection to delete from.
        ids: The IDs of the documents to delete.

    Returns:
        A PyMongo DeleteResult.
    """

    return await db.client.scooze[coll_type].delete_many({"_id": {"$in": ids}})


async def delete_documents(coll_type: DbCollection) -> DeleteResult:
    """
    Delete all documents in a single collection from the database.

    Args:
        coll_type: The collection to delete from.

    Returns:
        A PyMongo DeleteResult.
    """

    return await db.client.scooze[coll_type].delete_many({})  # NOTE: This deletes the entire collection.


# endregion



import argparse
import logging
import os.path
from datetime import date, datetime
from sys import maxsize, stdout
from typing import Any, Hashable, Iterable, Mapping, Self, Type, TypeVar

from frozendict import frozendict
from scooze.catalogs import ExtendedEnum, Format

DEFAULT_BULK_FILE_DIR = "./data/bulk"
DEFAULT_DECKS_DIR = "./data/decks"

## Generic Types
T = TypeVar("T")  # generic type
V = TypeVar("V")  # generic value type
E = TypeVar("E", bound=ExtendedEnum)  # generic Enum type
N = TypeVar("N", bound=ExtendedEnum)  # generic Enum (for mapping values) type
FloatableT = TypeVar("FloatableT", float, int, str)  # type that can normalize to float


def get_logger(
    filename: str,
    logger_name: str,
    file_logging_level: int = logging.DEBUG,
    console_logging_level: int = logging.WARNING,
    formatter: logging.Formatter = logging.Formatter("%(asctime)s - %(name)s:%(levelname)s - %(message)s"),
) -> logging.Logger:
    """
    Helper function to get a new logger.

    Args:
        filename: Filename of the log file.
        logger_name: The logger's name.
        file_logging_level: Logging level for the log file.
        console_logging_level: Logging level for stdout.
        formatter: The Formatter to be used in messages generated by this
        logger.

    Returns:
        A new logger.
    """

    # Create directory if not exists
    filepath = os.path.join("logs", filename)
    os.makedirs(os.path.dirname(filepath), exist_ok=True)

    # Handlers
    fh = logging.FileHandler(filepath, mode="a", encoding="UTF-8", delay=False)
    fh.setLevel(file_logging_level)
    ch = logging.StreamHandler(stdout)
    ch.setLevel(console_logging_level)

    # Formatting
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)

    # Create the logger
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(fh)
    logger.addHandler(ch)

    return logger


class SmartFormatter(argparse.RawDescriptionHelpFormatter, argparse.HelpFormatter):
    def _split_lines(self, text, width):
        if text.startswith("R|"):
            return text[2:].splitlines()
        # this is the RawTextHelpFormatter._split_lines
        return argparse.HelpFormatter._split_lines(self, text, width)


# region Deck Format Helpers


def max_relentless_quantity(name: str) -> int:
    """
    Given a card name, what is the maximum quantity of a card in a deck?
    """

    match name:
        case "Seven Dwarves":
            return 7
        case "Nazgûl" | "Nazgul":
            return 9
        case (
            "Plains"
            | "Island"
            | "Swamp"
            | "Mountain"
            | "Forest"
            | "Wastes"
            | "Snow-Covered Plains"
            | "Snow-Covered Island"
            | "Snow-Covered Swamp"
            | "Snow-Covered Mountain"
            | "Snow-Covered Forest"
            | "Dragon's Approach"
            | "Persistent Petitioners"
            | "Rat Colony"
            | "Relentless Rats"
            | "Shadowborn Apostle"
        ):
            return maxsize
        case _:
            return 0  # helps identify new relentless cards


def max_card_quantity(fmt: Format) -> int:
    """
    Given a Format, what is the maximum quantity of a card in a deck?
    """

    match fmt.value:
        case Format.LIMITED:
            return maxsize

        case (
            Format.BRAWL
            | Format.COMMANDER
            | Format.DUEL
            | Format.GLADIATOR
            | Format.HISTORICBRAWL
            | Format.OATHBREAKER
            | Format.PAUPERCOMMANDER
            | Format.PREDH
        ):
            return 1

        case (
            Format.ALCHEMY
            | Format.EXPLORER
            | Format.FUTURE
            | Format.HISTORIC
            | Format.LEGACY
            | Format.MODERN
            | Format.OLDSCHOOL
            | Format.PAUPER
            | Format.PENNY
            | Format.PIONEER
            | Format.PREMODERN
            | Format.STANDARD
            | Format.VINTAGE
        ):
            return 4

        case Format.NONE | _:
            return maxsize


def main_size(fmt: Format) -> tuple[int, int]:
    """
    Given a Format, what are the required min and max size for a main deck?
    """

    match fmt.value:
        case Format.LIMITED:
            return 40, maxsize

        case Format.OATHBREAKER:
            return 58, 58

        case (
            Format.ALCHEMY
            | Format.EXPLORER
            | Format.FUTURE
            | Format.HISTORIC
            | Format.LEGACY
            | Format.MODERN
            | Format.OLDSCHOOL
            | Format.PAUPER
            | Format.PENNY
            | Format.PIONEER
            | Format.PREMODERN
            | Format.STANDARD
            | Format.VINTAGE
        ):
            return 60, maxsize

        case (Format.BRAWL | Format.HISTORICBRAWL | Format.PAUPERCOMMANDER | Format.PREDH):
            return 99, 99

        case Format.COMMANDER | Format.DUEL:
            return 98, 99  # Accounting for Partner

        case Format.GLADIATOR:
            return 100, 100

        case Format.NONE | _:
            return 0, maxsize


def side_size(fmt: Format) -> tuple[int, int]:
    """
    Given a Format, what are the min and max size for a sideboard?
    """

    match fmt.value:
        case Format.LIMITED:
            return 0, maxsize

        case (
            Format.ALCHEMY
            | Format.EXPLORER
            | Format.FUTURE
            | Format.HISTORIC
            | Format.LEGACY
            | Format.MODERN
            | Format.OLDSCHOOL
            | Format.PAUPER
            | Format.PENNY
            | Format.PIONEER
            | Format.PREMODERN
            | Format.STANDARD
            | Format.VINTAGE
        ):
            return 0, 15

        case (
            Format.BRAWL
            | Format.COMMANDER
            | Format.DUEL
            | Format.GLADIATOR
            | Format.HISTORICBRAWL
            | Format.OATHBREAKER
            | Format.PAUPERCOMMANDER
            | Format.PREDH
        ):
            return 0, 0

        case Format.NONE | _:
            return 0, maxsize


def cmdr_size(fmt: Format) -> tuple[int, int]:
    """
    Given a Format, what are the min and max size for a command zone?
    """

    match fmt.value:
        case (
            Format.ALCHEMY
            | Format.EXPLORER
            | Format.FUTURE
            | Format.GLADIATOR
            | Format.HISTORIC
            | Format.LEGACY
            | Format.LIMITED
            | Format.MODERN
            | Format.OLDSCHOOL
            | Format.PAUPER
            | Format.PENNY
            | Format.PIONEER
            | Format.PREMODERN
            | Format.STANDARD
            | Format.VINTAGE
        ):
            return 0, 0

        case (Format.BRAWL | Format.HISTORICBRAWL | Format.PAUPERCOMMANDER | Format.PREDH):
            return 1, 1

        case Format.COMMANDER | Format.DUEL:
            return 1, 2  # Accounting for Partner

        case Format.OATHBREAKER:
            return 2, 2

        case Format.NONE | _:
            return 0, maxsize


# endregion


# region Helper Classes

# region Base Classes


class ComparableObject:
    """
    A simple base class to support comparable objects.
    """

    @property
    def __key__(self) -> tuple[Any, ...]:
        return tuple(getattr(self, k) for k in self.__dict__.keys())

    def __eq__(self, other: Self):
        return self.__key__ == other.__key__

    def __ne__(self, other: Self):
        return not (self == other)


class HashableObject(ComparableObject, Hashable):
    """
    A simple base class to support hashable objects.
    """

    def __hash__(self):
        return hash(self.__key__)


# endregion

# region JSON Utils


class JsonNormalizer:
    """
    A simple class to be used when normalizing non-serializable data from JSON.
    """

    @classmethod
    def to_date(cls, d: date | str | None) -> date:
        """
        Normalize a date.

        Args:
            d: A date to normalize.

        Returns:
            A date.
        """

        if d is None or isinstance(d, date):
            return d

        return datetime.strptime(d, "%Y-%m-%d").date()  # NOTE: maybe store date format

    @classmethod
    def to_enum(cls, e: Type[E], v) -> E | None:
        """
        Normalize an Enum.

        Args:
            e: A type of Enum to normalize to.
            v: A value to normalize.

        Returns:
            An instance of the given type of Enum.
        """

        if v is None:
            return v
        elif v in e.list():
            return e(v)

        return e[v]

    @classmethod
    def to_float(cls, f: FloatableT | None) -> float | None:
        """
        Normalize a float.

        Args:
            f: A float to normalize.

        Returns:
            A float.
        """

        if f is None or isinstance(f, float):
            return f

        return float(f)

    @classmethod
    def to_frozendict(
        cls, d: Mapping[T, V] | None, convert_key_to_enum: type[E] = None, convert_value_to_enum: type[N] = None
    ) -> frozendict[T | E, V | N] | None:
        """
        Normalize a frozendict.

        Args:
            d: A frozendict to normalize.
            convert_key_to_enum: A type of Enum to normalize keys to.
            convert_value_to_enum: A type of Enum to normalize values to.

        Returns:
            A frozendict.
        """

        if d is None:
            return d

        return frozendict(
            {
                JsonNormalizer.to_enum(e=convert_key_to_enum, v=k)
                if convert_key_to_enum
                else k: JsonNormalizer.to_enum(e=convert_value_to_enum, v=v)
                if convert_value_to_enum
                else v
                for k, v in d.items()
            }
        )

    @classmethod
    def to_frozenset(cls, s: Iterable[T] | None, convert_to_enum: type[E] = None) -> frozenset[T | E] | None:
        """
        Normalize a frozenset.

        Args:
            s: A frozenset to normalize.
            convert_to_enum: A type of Enum to normalize values to.

        Returns:
            A frozenset.
        """

        if s is None:
            return s

        return frozenset({JsonNormalizer.to_enum(e=convert_to_enum, v=v) if convert_to_enum else v for v in s})

    @classmethod
    def to_tuple(cls, t: Iterable[T] | None, convert_to_enum: type[E] = None) -> tuple[T | E] | None:
        """
        Normalize a tuple.

        Args:
            t: A tuple to normalize.
            convert_to_enum: A type of Enum to normalize values to.

        Returns:
            A tuple.
        """

        if t is None:
            return t

        return tuple([JsonNormalizer.to_enum(e=convert_to_enum, v=v) if convert_to_enum else v for v in t])


# endregion

# region Dict Diff


class DictDiff(ComparableObject):
    """
    Represents a diff between two dicts.

    Attributes:
        contents (dict[Any, tuple[int, int]]): The contents of this diff.
    """

    def __init__(self, contents: dict[T, tuple[int, int]]):
        self.contents = contents

    def __len__(self):
        return len(self.contents)

    def __str__(self):
        return "\n".join([f"{key}: {counts}" for key, counts in self.contents.items()]) + "\n"

    # Source:  https://code.activestate.com/recipes/576644-diff-two-dictionaries/#c9
    @classmethod
    def get_diff(cls, d1: dict, d2: dict, NO_KEY: Any = 0) -> "DictDiff":
        """
        Generate a diff between two dicts.

        Args:
            d1: The first dict.
            d2: The second dict.
            NO_KEY: Default value to use when a key is in one dict, but not the
              other.

        Returns:
            A dict with all keys from both dicts. The values are tuple(v, v)
            for the values in each dict.
        """

        both = d1.keys() & d2.keys()
        diff = {k: (d1[k], d2[k]) for k in both if d1[k] != d2[k]}
        diff.update({k: (d1[k], NO_KEY) for k in d1.keys() - both})
        diff.update({k: (NO_KEY, d2[k]) for k in d2.keys() - both})
        return DictDiff(diff)

    # endregion


# endregion



from datetime import date

from pydantic import Field, field_validator
from scooze.cardparts import (
    CardFace,
    FullCardFace,
    ImageUris,
    Preview,
    Prices,
    PurchaseUris,
    RelatedCard,
    RelatedUris,
)
from scooze.catalogs import (
    BorderColor,
    Color,
    Finish,
    Format,
    Frame,
    FrameEffect,
    Game,
    ImageStatus,
    Language,
    Layout,
    Legality,
    Rarity,
    SecurityStamp,
    SetType,
)
from scooze.models.cardparts import (
    CardFaceModel,
    ImageUrisModel,
    PreviewModel,
    PricesModel,
    PurchaseUrisModel,
    RelatedCardModel,
    RelatedUrisModel,
)
from scooze.models.utils import ObjectIdT, ScoozeBaseModel


class CardModel(ScoozeBaseModel):
    """
    Model for a basic Card object with minimal fields. Contains all information
      you might use to sort a decklist.

    Attributes:
        cmc: This card's mana value/converted mana cost.
        color_identity: This card's color identity, for Commander variant
          deckbuilding.
        colors: This card's colors.
        legalities: Formats and the legality status of this card in them.
        mana_cost: Mana cost, as string of mana symbols.
          (e.g. "{1}{W}{U}{B}{R}{G}")
        name: This card's name.
        power: Power of this card, if applicable.
        toughness: Toughness of this card, if applicable.
        type_line: This card's type line. (e.g. "Creature — Ooze")
    """

    model_config = ScoozeBaseModel.model_config.copy()
    model_config["json_schema_extra"] = {
        "examples": [
            {
                "cmc": 2.0,
                "colorIdentity": [Color.GREEN],
                "colors": [Color.GREEN],
                "legalities": {Format.COMMANDER: Legality.LEGAL, Format.PAUPER: Legality.NOT_LEGAL},
                "manaCost": "{1}{G}",
                "name": "Scavenging Ooze",
                "power": "2",
                "toughness": "2",
                "typeLine": "Creature — Ooze",
            }
        ]
    }

    cmc: float | None = Field(
        default=None,
        description="This card's mana value/converted mana cost.",
    )
    color_identity: set[Color] | None = Field(
        default=None,
        description="This card's color identity, for Commander variant deckbuilding.",
    )
    colors: set[Color] | None = Field(
        default=None,
        description="This card's colors.",
    )
    legalities: dict[Format, Legality] | None = Field(
        default=None,
        description="Formats and the legality status of this card in them.",
    )
    mana_cost: str | None = Field(
        default=None,
        description='Mana cost, as string of mana symbols. (e.g. "{1}{W}{U}{B}{R}{G}")',
    )
    name: str | None = Field(
        default=None,
        description="This card's name.",
    )
    power: str | None = Field(
        default=None,
        description="Power of this card, if applicable.",
    )
    toughness: str | None = Field(
        default=None,
        description="Toughness of this card, if applicable.",
    )
    type_line: str | None = Field(
        default=None,
        description='This card\'s type line. (e.g. "Creature — Ooze")',
    )

    # TODO(#46): add Card field validators


class FullCardModel(CardModel, validate_assignment=True):
    """
    Card object that supports all fields available from Scryfall's JSON data.
    Scryfall documentation: https://scryfall.com/docs/api/cards

    Attributes:
    Core fields
        arena_id: This card's Arena ID, if applicable.
        scryfall_id: Scryfall's unique ID for this card.
        lang: The language code for this print;
          see https://scryfall.com/docs/api/languages
        mtgo_id: This card's MTGO Catalog ID, if applicable.
        mtgo_foil_id: This card's foil MTGO Catalog ID, if applicable.
        multiverse_ids: This card's multiverse IDs on Gatherer, if any.
        tcgplayer_id: This card's ID on TCGplayer, or `productId` in their
          system.
        tcgplayer_etched_id: This card's ID on TCGplayer, for the etched
          version if that is a separate product.
        cardmarket_id: This card's ID on Cardmarket, or `idProduct` in their
          system.
        oracle_id: A UUID for this card's oracle identity; shared across prints
          of the same card but not same-named objects with different gameplay
          properties.
        prints_search_uri: A link to begin paginating through all prints of
          this card in Scryfall's API.
        rulings_uri: A link to rulings for this card in Scryfall's API.
        scryfall_uri: A link to the Scryfall page for this card.
        uri: A link to this card object in Scryfall's API.

    Gameplay fields
        all_parts: RelatedCard objects for tokens/meld pairs/other associated
          parts to this card, if applicable.
        card_faces: All component CardFace objects of this card, for multifaced
          cards.
        cmc: This card's mana value/converted mana cost.
        color_identity: This card's color identity, for Commander variant
          deckbuilding.
        color_indicator: color_indicator: The colors in this card's color
          indicator, if it has one.
        colors: This card's colors.
        edhrec_rank: This card's rank/popularity on EDHREC, if applicable.
        hand_modifier: This card's Vanguard hand size modifier, if applicable.
        keywords: Keywords and keyword actions this card uses.
        legalities: Formats and the legality status of this card in them.
        life_modifier: This card's Vanguard life modifier value, if applicable.
        loyalty: This card's starting planeswalker loyalty, if applicable.
        mana_cost: Mana cost, as string of mana symbols.
          (e.g. "{1}{W}{U}{B}{R}{G}")
        name: This card's name.
        oracle_text: This card's oracle text, if any.
        penny_rank: This card's rank/popularity on Penny Dreadful.
        power: Power of this card, if applicable.
        produced_mana: Which colors of mana this card can produce.
        reserved: Whether this card is on the Reserved List.
        toughness: Toughness of this card, if applicable.
        type_line: This card's type line. (e.g. "Creature — Ooze")

    Print fields
        artist: Artist for this card.
        artist_ids: List of Scryfall IDs for artists of this card.
        attraction_lights: Attraction lights lit on this card, if applicable.
        booster: Whether this card can be opened in booster packs.
        border_color: Border color of this card, from among
          black, white, borderless, silver, and gold.
        card_back_id: Scryfall UUID of the card back design for this card.
        collector_number: This card's collector number; can contain non-numeric
          characters.
        content_warning: True if use of this print should be avoided;
          see https://scryfall.com/blog/regarding-wotc-s-recent-statement-on-depictions-of-racism-220
        digital: True if this card was only released in a video game.
        finishes: Finishes this card is available in, from among foil, nonfoil, and etched.
        flavor_name: Alternate name for this card, such as on Godzilla series.
        flavor_text: Flavor text on this card, if any.
        frame_effects: Special frame effects on this card;
          see https://scryfall.com/docs/api/frames
        frame: This card's frame layout;
          see https://scryfall.com/docs/api/frames
        full_art: Whether this print is full-art.
        games: Which games this print is available on, from among
          paper, mtgo, and arena.
        highres_image: Whether this card has a high-res image available.
        illustration_id: A UUID for the particular artwork on this print,
          consistent across art reprints.
        image_status: The quality/status of images available for this card.
          Either missing, placeholder, lowres, or highres_scan.
        image_uris: Links to images of this card in various qualities.
        layout: This card's printed layout;
          see https://scryfall.com/docs/api/layouts
        oversized: Whether this card is oversized.
        preview: Information about where, when, and how this print was
          previewed.
        prices: Prices for this card on various marketplaces.
        printed_name: Printed name of this card, for localized non-English
          cards.
        printed_text: Printed text of this card, for localized non-English
          cards.
        printed_type_line: Printed type line of this card, for localized
          non-English cards.
        promo: Whether this print is a promo.
        promo_types: Which promo categories this print falls into, if any.
        purchase_uris: Links to purchase this print from marketplaces.
        rarity: The rarity of this print.
        related_uris: Links to this print's listing on other online resources.
        released_at: The date this card was first released.
        reprint: Whether this print is a reprint from an earlier set.
        scryfall_set_uri: Link to the Scryfall set page for the set of this
          print.
        security_stamp: Security stamp on this card, if any.
        set_name: Full name of the set this print belongs to.
        set_search_uri: Link to Scryfall API to start paginating through this
          print's full set.
        set_type: An overall categorization for each set, provided by Scryfall.
        set_uri: Link to the set object for this print in Scryfall's API.
        set: Set code of the set this print belongs to.
        set_id: UUID of the set this print belongs to.
        story_spotlight: Whether this print is a Story Spotlight.
        textless: Whether this print is textless.
        variation: Whether this card print is a variation of another card
          object.
        variation_of: Which card object this object is a variant of, if any.
        watermark: Watermark printed on this card, if any.
    """

    # region Core fields

    arena_id: int | None = Field(
        default=None,
        description="This card's Arena ID, if applicable.",
    )
    scryfall_id: str | None = Field(
        default=None,
        description="Scryfall's unique ID for this card.",
        validation_alias="id",
    )
    lang: Language | None = Field(
        default=None,
        description="The language code for this print; see https://scryfall.com/docs/api/languages",
    )
    mtgo_id: int | None = Field(
        default=None,
        description="This card's MTGO Catalog ID, if applicable.",
    )
    mtgo_foil_id: int | None = Field(
        default=None,
        description="This card's foil MTGO Catalog ID, if applicable.",
    )
    multiverse_ids: list[int] | None = Field(
        default=None,
        description="This card's multiverse IDs on Gatherer, if any.",
    )
    tcgplayer_id: int | None = Field(
        default=None,
        description="This card's ID on TCGplayer, or `productId` in their system.",
    )
    tcgplayer_etched_id: int | None = Field(
        default=None,
        description="This card's ID on TCGplayer, for the etched version if that is a separate product.",
    )
    cardmarket_id: int | None = Field(
        default=None,
        description="This card's ID on Cardmarket, or `idProduct` in their system.",
    )
    oracle_id: str | None = Field(
        default=None,
        description="A UUID for this card's oracle identity; shared across prints of the same card but not same-named objects with different gameplay properties.",
    )
    prints_search_uri: str | None = Field(
        default=None,
        description="A link to begin paginating through all prints of this card in Scryfall's API.",
    )
    rulings_uri: str | None = Field(
        default=None,
        description="A link to rulings for this card in Scryfall's API.",
    )
    scryfall_uri: str | None = Field(
        default=None,
        description="A link to the Scryfall page for this card.",
    )
    uri: str | None = Field(
        default=None,
        description="A link to this card object in Scryfall's API.",
    )

    # endregion

    # region Gameplay fields

    all_parts: list[RelatedCardModel] | None = Field(
        default=None,
        description="RelatedCard objects for tokens/meld pairs/other associated parts to this card, if applicable.",
    )
    card_faces: list[CardFaceModel] | None = Field(
        default=None,
        description="All component CardFace objects of this card, for multifaced cards.",
    )
    # cmc defined by base model
    # color_identity defined by base model
    color_indicator: set[Color] | None = Field(
        default=None,
        description="The colors in this card's color indicator, if it has one.",
    )
    # colors defined by base model
    edhrec_rank: int | None = Field(
        default=None,
        description="This card's rank/popularity on EDHREC, if applicable.",
    )
    hand_modifier: str | None = Field(
        default=None,
        description="This card's Vanguard hand size modifier, if applicable.",
    )
    keywords: set[str] | None = Field(
        default=None,
        description="Keywords and keyword actions this card uses.",
    )
    # legalities defined by base model
    life_modifier: str | None = Field(
        default=None,
        description="This card's Vanguard life modifier value, if applicable.",
    )
    loyalty: str | None = Field(
        default=None,
        description="This card's starting planeswalker loyalty, if applicable.",
    )
    # mana_cost defined by base model
    # name defined by base model
    oracle_text: str | None = Field(
        default=None,
        description="This card's oracle text, if any.",
    )
    penny_rank: int | None = Field(
        default=None,
        description="This card's rank/popularity on Penny Dreadful.",
    )
    # power defined by base model
    produced_mana: set[Color] | None = Field(
        default=None,
        description="Which colors of mana this card can produce.",
    )
    reserved: bool | None = Field(
        default=None,
        description="Whether this card is on the Reserved List.",
    )
    # toughness defined by base model
    # type_line defined by base model

    # endregion

    # region Print fields

    artist: str | None = Field(
        default=None,
        description="Artist for this card.",
    )
    artist_ids: list[str] | None = Field(
        default=None,
        description="List of Scryfall IDs for artists of this card.",
    )
    attraction_lights: set[int] | None = Field(
        default=None,
        description="Attraction lights lit on this card, if applicable.",
    )
    booster: bool | None = Field(
        default=None,
        description="Whether this card can be opened in booster packs.",
    )
    border_color: BorderColor | None = Field(
        default=None,
        description="Border color of this card, from among black, white, borderless, silver, and gold.",
    )
    card_back_id: str | None = Field(
        default=None,
        description="Scryfall UUID of the card back design for this card.",
    )
    collector_number: str | None = Field(
        default=None,
        description="This card's collector number; can contain non-numeric characters.",
    )
    content_warning: bool = Field(
        default=False,
        description="True if use of this print should be avoided; see https://scryfall.com/blog/regarding-wotc-s-recent-statement-on-depictions-of-racism-220",
    )
    digital: bool | None = Field(
        default=None,
        description="True if this card was only released in a video game.",
    )
    finishes: set[Finish] | None = Field(
        default=None,
        description="Finishes this card is available in, from among foil, nonfoil, and etched.",
    )
    flavor_name: str | None = Field(
        default=None,
        description="Alternate name for this card, such as on Godzilla series.",
    )
    flavor_text: str | None = Field(
        default=None,
        description="Flavor text on this card, if any.",
    )
    frame_effects: set[FrameEffect] | None = Field(
        default=None,
        description="Special frame effects on this card; see https://scryfall.com/docs/api/frames",
    )
    frame: Frame | None = Field(
        default=None,
        description="This card's frame layout; see https://scryfall.com/docs/api/frames",
    )
    full_art: bool | None = Field(
        default=None,
        description="Whether this print is full-art.",
    )
    games: set[Game] | None = Field(
        default=None,
        description="Which games this print is available on, from among paper, mtgo, and arena.",
    )
    highres_image: bool | None = Field(
        default=None,
        description="Whether this card has a high-res image available.",
    )
    illustration_id: str | None = Field(
        default=None,
        description="A UUID for the particular artwork on this print, consistent across art reprints.",
    )
    image_status: ImageStatus | None = Field(
        default=None,
        description="The quality/status of images available for this card. Either missing, placeholder, lowres, or highres_scan.",
    )
    image_uris: ImageUrisModel | None = Field(
        default=None,
        description="Links to images of this card in various qualities.",
    )
    layout: Layout | None = Field(
        default=None,
        description="This card's printed layout; see https://scryfall.com/docs/api/layouts",
    )
    oversized: bool | None = Field(
        default=None,
        description="Whether this card is oversized.",
    )
    preview: PreviewModel | None = Field(
        default=None,
        description="Information about where, when, and how this print was previewed.",
    )
    prices: PricesModel | None = Field(
        default=None,
        description="Prices for this card on various marketplaces.",
    )
    printed_name: str | None = Field(
        default=None,
        description="Printed name of this card, for localized non-English cards.",
    )
    printed_text: str | None = Field(
        default=None,
        description="Printed text of this card, for localized non-English cards.",
    )
    printed_type_line: str | None = Field(
        default=None,
        description="Printed type line of this card, for localized non-English cards.",
    )
    promo: bool | None = Field(
        default=None,
        description="Whether this print is a promo.",
    )
    promo_types: set[str] | None = Field(
        default=None,
        description="Which promo categories this print falls into, if any.",
    )
    purchase_uris: PurchaseUrisModel | None = Field(
        default=None,
        description="Links to purchase this print from marketplaces.",
    )
    rarity: Rarity | None = Field(
        default=None,
        description="The rarity of this print.",
    )
    related_uris: RelatedUrisModel | None = Field(
        default=None,
        description="Links to this print's listing on other online resources.",
    )
    released_at: date | None = Field(
        default=None,
        description="The date this card was first released.",
    )
    reprint: bool | None = Field(
        default=None,
        description="Whether this print is a reprint from an earlier set.",
    )
    scryfall_set_uri: str | None = Field(
        default=None,
        description="Link to the Scryfall set page for the set of this print.",
    )
    security_stamp: SecurityStamp | None = Field(
        default=None,
        description="Security stamp on this card, if any.",
    )
    set_name: str | None = Field(
        default=None,
        description="Full name of the set this print belongs to.",
    )
    set_search_uri: str | None = Field(
        default=None,
        description="Link to Scryfall API to start paginating through this print's full set.",
    )
    set_type: SetType | None = Field(
        default=None,
        description="An overall categorization for each set, provided by Scryfall.",
    )
    set_uri: str | None = Field(
        default=None,
        description="Link to the set object for this print in Scryfall's API.",
    )
    set: str | None = Field(
        default=None,
        description="Set code of the set this print belongs to.",
    )
    set_id: str | None = Field(
        default=None,
        description="UUID of the set this print belongs to.",
    )
    story_spotlight: bool | None = Field(
        default=None,
        description="Whether this print is a Story Spotlight.",
    )
    textless: bool | None = Field(
        default=None,
        description="Whether this print is textless.",
    )
    variation: bool | None = Field(
        default=None,
        description="Whether this card print is a variation of another card object.",
    )
    variation_of: str | None = Field(
        default=None,
        description="Which card object this object is a variant of, if any.",
    )
    watermark: str | None = Field(
        default=None,
        description="Watermark printed on this card, if any.",
    )

    # endregion

    # region Validators

    @field_validator("all_parts", mode="before")
    @classmethod
    def all_parts_validator(cls, vals):
        if vals is not None and all(isinstance(v, RelatedCard) for v in vals):
            vals = [RelatedCardModel.model_validate(v.__dict__) for v in vals]
        return vals

    @field_validator("card_faces", mode="before")
    @classmethod
    def card_faces_validator(cls, vals):
        if vals is not None and all(isinstance(v, CardFace | FullCardFace) for v in vals):
            vals = [CardFaceModel.model_validate(v.__dict__) for v in vals]
        return vals

    @field_validator("image_uris", mode="before")
    @classmethod
    def image_uris_validator(cls, v):
        if isinstance(v, ImageUris):
            v = ImageUrisModel.model_validate(v.__dict__)
        return v

    @field_validator("preview", mode="before")
    @classmethod
    def preview_validator(cls, v):
        if isinstance(v, Preview):
            v = PreviewModel.model_validate(v.__dict__)
        return v

    @field_validator("prices", mode="before")
    @classmethod
    def prices_validator(cls, v):
        if isinstance(v, Prices):
            v = PricesModel.model_validate(v.__dict__)
        return v

    @field_validator("purchase_uris", mode="before")
    @classmethod
    def purchase_uris_validator(cls, v):
        if isinstance(v, PurchaseUris):
            v = PurchaseUrisModel.model_validate(v.__dict__)
        return v

    @field_validator("related_uris", mode="before")
    @classmethod
    def related_uris_validator(cls, v):
        if isinstance(v, RelatedUris):
            v = RelatedUrisModel.model_validate(v.__dict__)
        return v

    # endregion


class CardModelIn(FullCardModel):
    pass


class CardModelOut(FullCardModel):
    scooze_id: ObjectIdT = Field(
        default=None,
        alias="_id",
    )



import subprocess

import docker
from cleo.commands.command import Command


class SetupDockerCommand(Command):
    name = "setup docker"
    description = "Setup MongoDB in a Docker container."

    def handle(self):
        # Check if Docker is installed and running
        p = subprocess.run(
            "docker stats --no-stream",
            stdout=subprocess.DEVNULL,
            stderr=subprocess.STDOUT,
            shell=True,
        )
        if not p.returncode:
            client = docker.from_env()
            # Check if Docker container is already running
            containers = client.containers.list(all=True)
            if "scooze-mongodb" in [container.name for container in containers]:
                self.line("Scooze mongodb container already exists! Exiting.")
            else:
                self.line("Setting up latest MongoDB Docker container as scooze-mongodb...")
                # Start Docker container
                client.containers.run("mongo:latest", detach=True, ports=({"27017/tcp": 27017}), name="scooze-mongodb")
                self.line("Done. MongoDB running on localhost:27017.")
        else:
            self.line("Cannot connect to Docker daemon -- Is docker installed and running?")



from typing import Any, List

import scooze.database.card as db
from bson import ObjectId
from scooze.card import CardT, FullCard
from scooze.errors import BulkAddError
from scooze.models.card import CardModelIn


async def get_card_by(property_name: str, value, card_class: CardT = FullCard) -> CardT:
    """
    Search the database for the first card that matches the given criteria.

    Args:
        property_name: The property to check.
        value: The value to match on.
        card_class: The type of card to return.

    Returns:
        The first matching card, or None if none were found.
    """

    card_model = await db.get_card_by_property(
        property_name=property_name,
        value=value,
    )
    if card_model is not None:
        return card_class.from_model(card_model)


async def get_cards_by(
    property_name: str,
    values: list[Any],
    card_class: CardT = FullCard,
    paginated: bool = False,
    page: int = 1,
    page_size: int = 10,
) -> List[CardT]:
    """
    Search the database for cards matching the given criteria, with options for
    pagination.

    Args:
        property_name: The property to check.
        values: A list of values to match on.
        card_class: The type of card object to return.
        paginated: Whether to paginate the results.
        page: The page to look at, if paginated.
        page_size: The size of each page, if paginated.

    Returns:
        A list of cards matching the search criteria, or empty list if none
        were found.
    """

    card_models = await db.get_cards_by_property(
        property_name=property_name,
        values=values,
        paginated=paginated,
        page=page,
        page_size=page_size,
    )
    return [card_class.from_model(m) for m in card_models]


async def get_cards_all(
    card_class: CardT = FullCard,
) -> List[CardT]:
    """
    Get all cards from the database. WARNING: may be extremely large.

    Returns:
        A list of all cards in the database.
    """

    card_models = await db.get_cards_all()
    return [card_class.from_model(m) for m in card_models]


async def add_card(card: CardT) -> ObjectId:
    """
    Add a card to the database.

    Assign the resulting database ID to the given Card.

    Args:
        card: The card to insert.

    Returns:
        The ID of the inserted card, or None if it was unable.
    """

    card_model = CardModelIn.model_validate(card.__dict__)
    model = await db.add_card(card=card_model)

    if model is not None:
        card.scooze_id = model.scooze_id
        return model.scooze_id


async def add_cards(cards: List[CardT]) -> List[ObjectId]:
    """
    Add a list of cards to the database.

    Assign the resulting database IDs to the given Cards.

    Args:
        cards: The list of cards to insert.

    Returns:
        The IDs of the inserted cards, or empty list if no cards provided.

    Raises:
        BulkAddError: If not all IDs are successfully inserted.
    """

    if not cards:
        return []

    card_models = [CardModelIn.model_validate(card.__dict__) for card in cards]
    card_ids = await db.add_cards(cards=card_models)

    if len(card_ids) != len(cards):
        # TODO(#202): Perform card lookups to get the ids of the cards that were successfully added.
        await db.delete_cards_by_id(card_ids)
        raise BulkAddError("Failed to add all cards to the database.")
    else:
        for card, card_id in zip(cards, card_ids):
            card.scooze_id = card_id

    return card_ids


async def delete_card(id: str) -> bool:
    """
    Delete a card from the database.

    Args:
        id: The ID of the card to delete.

    Returns:
        True if the card is deleted, False otherwise.
    """

    return await db.delete_card(id=id) is not None


async def delete_cards_all() -> int:
    """
    Delete all cards in the database.

    Returns:
        The number of cards deleted, or None if none could be deleted.
    """

    return await db.delete_cards_all()



import asyncio
from contextlib import AbstractAsyncContextManager, AbstractContextManager
from functools import cache
from typing import Any, List

import scooze.api.bulkdata as bulkdata_api
import scooze.api.card as card_api
import scooze.database.mongo as mongo
from bson import ObjectId
from scooze.api.utils import _check_for_safe_context, _safe_cache
from scooze.card import CardT, FullCard
from scooze.catalogs import ScryfallBulkFile


class ScoozeApi(AbstractContextManager):
    """
    Context manager object for doing I/O from a Mongo database.

    Sample usage:
        >>> with ScoozeApi() as s:
                green_cards = s.get_cards_by("colors", [Color.GREEN])
                woe_cards = s.get_cards_by_set("woe")
                black_lotus = s.get_card_by_scryfall_id("b0faa7f2-b547-42c4-a810-839da50dadfe")
                print(black_lotus.total_words())
    """

    def __init__(self, card_class: type[CardT] = FullCard):
        self.card_class = card_class
        self.safe_context = False

    def __enter__(self):
        self.safe_context = True
        self.runner = asyncio.Runner()
        self.runner.run(mongo.mongo_connect())

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.runner.run(mongo.mongo_close())

    # region Card endpoints

    @_safe_cache
    @_check_for_safe_context
    def get_card_by(self, property_name: str, value) -> CardT:
        """
        Search the database for the first card that matches the given criteria.

        Args:
            property_name: The property to check.
            value: The value to match on.

        Returns:
            The first matching card, or None if none were found.

        Raises:
            RuntimeError: If used outside a `with` context.
        """

        return self.runner.run(
            card_api.get_card_by(property_name=property_name, value=value, card_class=self.card_class)
        )

    @_check_for_safe_context
    def get_cards_by(
        self,
        property_name: str,
        values: list[Any],
        paginated: bool = False,
        page: int = 1,
        page_size: int = 10,
    ) -> List[CardT]:
        """
        Search the database for cards matching the given criteria, with options for
        pagination.

        Args:
            property_name: The property to check.
            values: A list of values to match on.
            paginated: Whether to paginate the results.
            page: The page to look at, if paginated.
            page_size: The size of each page, if paginated.

        Returns:
            A list of cards matching the search criteria, or empty list if none
            were found.

        Raises:
            RuntimeError: If used outside a `with` context.
        """

        return self.runner.run(
            card_api.get_cards_by(
                property_name=property_name,
                values=values,
                card_class=self.card_class,
                paginated=paginated,
                page=page,
                page_size=page_size,
            )
        )

    # region Convenience methods for single-card lookup

    @cache
    @_check_for_safe_context
    def get_card_by_name(self, name: str) -> CardT:
        """
        Search the database for a card with the given name.

        Args:
            name: The card name to search for.

        Returns:
            A card with the given name if found, or None if none were found.

        Raises:
            RuntimeError: If used outside a `with` context.
        """

        return self.runner.run(
            card_api.get_card_by(
                property_name="name",
                value=name,
                card_class=self.card_class,
            )
        )

    @cache
    @_check_for_safe_context
    def get_card_by_oracle_id(self, oracle_id: str) -> CardT:
        """
        Search the database for a card with the given Oracle ID.

        Args:
            oracle_id: The card [Oracle ID](https://scryfall.com/docs/api/cards) to search for.

        Returns:
            A card with the given Oracle ID if found, or None if none were found.

        Raises:
            RuntimeError: If used outside a `with` context.
        """

        return self.runner.run(
            card_api.get_card_by(
                property_name="oracle_id",
                value=oracle_id,
                card_class=self.card_class,
            )
        )

    @cache
    @_check_for_safe_context
    def get_card_by_scryfall_id(self, scryfall_id: str) -> CardT:
        """
        Search the database for a card with the given Scryfall ID.

        Args:
            scryfall_id: The card [Scryfall ID](https://scryfall.com/docs/api/cards) to search for.

        Returns:
            A card with the given Scryfall ID if found, or None if none were found.

        Raises:
            RuntimeError: If used outside a `with` context.
        """

        return self.runner.run(
            card_api.get_card_by(
                property_name="scryfall_id",
                value=scryfall_id,
                card_class=self.card_class,
            )
        )

    # endregion

    # region Convenience methods for multiple card lookup

    @_check_for_safe_context
    def get_cards_by_set(self, set_code: str) -> List[CardT]:
        """
         Search the database for all cards in the given set.
         Expects the 3-letter [set code](https://en.wikipedia.org/wiki/List_of_Magic:_The_Gathering_sets)
         for a set (e.g. "CMD")

         Args:
             set_code: The set code to search for.

         Returns:
            A list of cards from the given set, or empty list if none were found.

        Raises:
             RuntimeError: If used outside a `with` context.
        """

        return self.runner.run(
            card_api.get_cards_by(
                property_name="set",
                values=[set_code],
                card_class=self.card_class,
            )
        )

    @_check_for_safe_context
    def get_cards_all(self) -> List[CardT]:
        """
        Get all cards from the database. WARNING: may be extremely large.

        Returns:
            A list of all cards in the database.

        Raises:
            RuntimeError: If used outside a `with` context.
        """

        return self.runner.run(card_api.get_cards_all(self.card_class))

    # TODO(#146): add function get_cards_by_format (format, legality)

    # endregion

    @_check_for_safe_context
    def add_card(self, card: CardT) -> ObjectId:
        """
        Add a card to the database.

        Args:
            card: The card to insert.

        Returns:
            The ID of the inserted card, or None if it was unable.

        Raises:
            RuntimeError: If used outside a `with` context.
        """

        return self.runner.run(card_api.add_card(card=card))

    @_check_for_safe_context
    def add_cards(self, cards: List[CardT]) -> List[ObjectId]:
        """
        Add a list of cards to the database.

        Args:
            cards: The list of card to insert.

        Returns:
            The IDs of the inserted cards, or empty list if unable.

        Raises:
            RuntimeError: If used outside a `with` context.
        """

        return self.runner.run(card_api.add_cards(cards=cards))

    @_check_for_safe_context
    def delete_card(self, id: str) -> bool:
        """
        Delete a card from the database.

        Args:
            id: The ID of the card to delete.

        Returns:
            True if the card is deleted, False otherwise.

        Raises:
            RuntimeError: If used outside a `with` context.
        """

        return self.runner.run(card_api.delete_card(id=id))

    @_check_for_safe_context
    def delete_cards_all(self) -> int:
        """
        Delete all cards in the database.

        Returns:
            The number of cards deleted, or None if none could be deleted.

        Raises:
            RuntimeError: If used outside a `with` context.
        """

        return self.runner.run(card_api.delete_cards_all())

    # endregion

    # region Deck endpoints

    # TODO(#145): add deck endpoints to python api

    # endregion

    # region Bulk data I/O

    @_check_for_safe_context
    def load_card_file(self, file_type: ScryfallBulkFile, bulk_file_dir: str):
        """
        Loads the desired file from the given directory into a local Mongo
        database. Attempts to download it from Scryfall if it isn't found.

        Args:
            file_type: The type of [ScryfallBulkFile](https://scryfall.com/docs/api/bulk-data)
            to insert into the database.
            bulk_file_dir: The path to the folder containing the ScryfallBulkFile.

        Raises:
            RuntimeError: If used outside a `with` context.
        """

        return self.runner.run(
            bulkdata_api.load_card_file(
                file_type=file_type,
                bulk_file_dir=bulk_file_dir,
            )
        )

    # endregion


class AsyncScoozeApi(AbstractAsyncContextManager):
    """
    Async context manager object for doing I/O from a Mongo database.
    Most commonly used in asynchronous contexts like Jupyter Notebooks or other
    web applications.

    Sample usage:
        >>> async with AsyncScoozeApi() as s:
                green_cards = await s.get_cards_by("colors", [Color.GREEN])
                woe_cards = await s.get_cards_by_set("woe")
                black_lotus = await s.get_card_by_scryfall_id("b0faa7f2-b547-42c4-a810-839da50dadfe")
                print(black_lotus.total_words())
    """

    def __init__(self, card_class: type[CardT] = FullCard):
        self.card_class = card_class
        self.safe_context = False

    async def __aenter__(self):
        self.safe_context = True
        await mongo.mongo_connect()

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await mongo.mongo_close()

    # region Card endpoints

    @_safe_cache
    @_check_for_safe_context
    async def get_card_by(self, property_name: str, value) -> CardT:
        """
        Search the database for the first card that matches the given criteria.

        Args:
            property_name: The property to check.
            value: The value to match on.

        Returns:
            The first matching card, or None if none were found.

        Raises:
            RuntimeError: If used outside an `async with` context.
        """

        return await card_api.get_card_by(property_name=property_name, value=value, card_class=self.card_class)

    @_check_for_safe_context
    async def get_cards_by(
        self,
        property_name: str,
        values: list[Any],
        paginated: bool = False,
        page: int = 1,
        page_size: int = 10,
    ) -> List[CardT]:
        """
        Search the database for cards matching the given criteria, with options for
        pagination.

        Args:
            property_name: The property to check.
            values: A list of values to match on.
            paginated: Whether to paginate the results.
            page: The page to look at, if paginated.
            page_size: The size of each page, if paginated.

        Returns:
            A list of cards matching the search criteria, or empty list if none
            were found.

        Raises:
            RuntimeError: If used outside an `async with` context.
        """

        return await card_api.get_cards_by(
            property_name=property_name,
            values=values,
            card_class=self.card_class,
            paginated=paginated,
            page=page,
            page_size=page_size,
        )

    # region Convenience methods for single-card lookup

    @cache
    @_check_for_safe_context
    async def get_card_by_name(self, name: str) -> CardT:
        """
        Search the database for a card with the given name.

        Args:
            name: The card name to search for.

        Returns:
            A card with the given name if found, or None if none were found.

        Raises:
            RuntimeError: If used outside an `async with` context.
        """

        return await card_api.get_card_by(
            property_name="name",
            value=name,
            card_class=self.card_class,
        )

    @cache
    @_check_for_safe_context
    async def get_card_by_oracle_id(self, oracle_id: str) -> CardT:
        """
        Search the database for a card with the given Oracle ID.

        Args:
            oracle_id: The card [Oracle ID](https://scryfall.com/docs/api/cards) to search for.

        Returns:
            A card with the given Oracle ID if found, or None if none were found.

        Raises:
            RuntimeError: If used outside an `async with` context.
        """

        return await card_api.get_card_by(
            property_name="oracle_id",
            value=oracle_id,
            card_class=self.card_class,
        )

    @cache
    @_check_for_safe_context
    async def get_card_by_scryfall_id(self, scryfall_id: str) -> CardT:
        """
        Search the database for a card with the given Scryfall ID.

        Args:
            scryfall_id: The card [Scryfall ID](https://scryfall.com/docs/api/cards) to search for.

        Returns:
            A card with the given Scryfall ID if found, or None if none were found.

        Raises:
            RuntimeError: If used outside an `async with` context.
        """

        return await card_api.get_card_by(
            property_name="scryfall_id",
            value=scryfall_id,
            card_class=self.card_class,
        )

    # endregion

    # region Convenience methods for multiple card lookup

    @_check_for_safe_context
    async def get_cards_by_set(self, set_code: str) -> List[CardT]:
        """
        Search the database for all cards in the given set.
        Expects the 3-letter [set code](https://en.wikipedia.org/wiki/List_of_Magic:_The_Gathering_sets)
        for a set (e.g. "CMD")

        Args:
            set_code: The set code to search for.

        Returns:
            A list of cards from the given set, or empty list if none were found.

        Raises:
            RuntimeError: If used outside an `async with` context.
        """

        return await card_api.get_cards_by(
            property_name="set",
            values=[set_code],
            card_class=self.card_class,
        )

    @_check_for_safe_context
    async def get_cards_all(self) -> List[CardT]:
        """
        Get all cards from the database. WARNING: may be extremely large.

        Returns:
            A list of all cards in the database.

        Raises:
            RuntimeError: If used outside an `async with` context.
        """

        return await card_api.get_cards_all(self.card_class)

    # TODO(#146): add function get_cards_by_format (format, legality)

    # endregion

    @_check_for_safe_context
    async def add_card(self, card: CardT) -> ObjectId:
        """
        Add a card to the database.

        Args:
            card: The card to insert.

        Returns:
            The ID of the inserted card, or None if it was unable.

        Raises:
            RuntimeError: If used outside an `async with` context.
        """

        return await card_api.add_card(card=card)

    @_check_for_safe_context
    async def add_cards(self, cards: List[CardT]) -> List[ObjectId]:
        """
        Add a list of cards to the database.

        Args:
            cards: The list of card to insert.

        Returns:
            The IDs of the inserted cards, or empty list if unable.

        Raises:
            RuntimeError: If used outside an `async with` context.
        """

        return await card_api.add_cards(cards=cards)

    @_check_for_safe_context
    async def delete_card(self, id: str) -> bool:
        """
        Delete a card from the database.

        Args:
            id: The ID of the card to delete.

        Returns:
            True if the card is deleted, False otherwise.

        Raises:
            RuntimeError: If used outside an `async with` context.
        """

        return await card_api.delete_card(id=id)

    @_check_for_safe_context
    async def delete_cards_all(self) -> int:
        """
        Delete all cards in the database.

        Returns:
            The number of cards deleted, or None if none could be deleted.

        Raises:
            RuntimeError: If used outside an `async with` context.
        """

        return await card_api.delete_cards_all()

    # endregion

    # region Deck endpoints

    # TODO(#145): add deck endpoints to python api

    # endregion

    # region Bulk data I/O

    @_check_for_safe_context
    async def load_card_file(self, file_type: ScryfallBulkFile, bulk_file_dir: str):
        """
        Loads the desired file from the given directory into a local Mongo
        database. Attempts to download it from Scryfall if it isn't found.

        Args:
            file_type: The type of [ScryfallBulkFile](https://scryfall.com/docs/api/bulk-data)
            to insert into the database.
            bulk_file_dir: The path to the folder containing the ScryfallBulkFile.

        Raises:
            RuntimeError: If used outside an `async with` context.
        """

        return await bulkdata_api.load_card_file(
            file_type=file_type,
            bulk_file_dir=bulk_file_dir,
        )

    # endregion



from typing import Any

import scooze.database.core as db_core
from bson import ObjectId
from scooze.catalogs import DbCollection
from scooze.models.card import CardModelIn, CardModelOut

# region Card


async def add_card(card: CardModelIn) -> CardModelOut:
    """
    Add a card to the database.

    Args:
        card: The card to insert.

    Returns:
        The card that was inserted, or None if it was unable.
    """

    new_card = await db_core.insert_document(
        DbCollection.CARDS,
        card.model_dump(
            mode="json",
            by_alias=True,
        ),
    )
    if new_card is not None:
        return CardModelOut.model_validate(new_card)


async def get_card_by_property(property_name: str, value) -> CardModelOut:
    """
    Search the database for the first card that matches the given criteria.

    Args:
        property_name: The property to check.
        value: The value to match on.

    Returns:
        The first matching card, or None if none were found.
    """

    card = await db_core.get_document_by_property(DbCollection.CARDS, property_name, value)
    if card is not None:
        return CardModelOut.model_validate(card)


async def update_card(id: str, card: CardModelIn) -> CardModelOut:
    """
    Update a card in the database with the given values.

    Args:
        id: The ID of the card to update.
        card: The properties to update and their new values.

    Returns:
        The updated card, or None if it was unable to update or find it.
    """

    updated_card = await db_core.update_document(
        DbCollection.CARDS,
        id,
        card.model_dump(
            mode="json",
            by_alias=True,
            include=card.model_fields_set,
        ),
    )
    if updated_card is not None:
        return CardModelOut.model_validate(updated_card)


async def delete_card(id: str) -> CardModelOut:
    """
    Delete a card from the database.

    Args:
        id: The ID of the card to delete.

    Returns:
        The deleted card, or None if unable to delete or find it.
    """

    deleted_card = await db_core.delete_document(DbCollection.CARDS, id)

    if deleted_card is not None:
        return CardModelOut.model_validate(deleted_card)


# endregion


# region Cards


async def add_cards(cards: list[CardModelIn]) -> list[ObjectId]:
    """
    Add a list of cards to the database.

    Args:
        cards: The list of card to insert.

    Returns:
        The list of IDs for cards that were inserted, or empty list if unable.
    """

    insert_many_result = await db_core.insert_many_documents(
        DbCollection.CARDS,
        [
            card.model_dump(
                mode="json",
                by_alias=True,
            )
            for card in cards
        ],
    )
    return insert_many_result.inserted_ids if insert_many_result is not None else []


async def get_cards_random(limit: int) -> list[CardModelOut]:
    """
    Get a random assortment of cards from the database.

    Args:
        limit: The number of cards to return.

    Returns:
        A random list of cards, or empty list if none were found.
    """

    cards = await db_core.get_random_documents(DbCollection.CARDS, limit)
    return [CardModelOut.model_validate(card) for card in cards]


async def get_cards_all() -> list[CardModelOut]:
    """
    Get all cards from the database. WARNING: may be extremely large.

    Returns:
        A list of all cards in the database.
    """

    cards = await db_core.get_all_documents(DbCollection.CARDS)
    return [CardModelOut.model_validate(card) for card in cards]


async def get_cards_by_property(
    property_name: str, values: list[Any], paginated: bool = False, page: int = 1, page_size: int = 10
) -> list[CardModelOut]:
    """
    Search the database for cards matching the given criteria, with options for
    pagination.

    Args:
        property_name: The property to check.
        values: A list of values to match on.
        paginated: Whether to paginate the results.
        page: The page to look at, if paginated.
        page_size: The size of each page, if paginated.

    Returns:
        A list of cards matching the search criteria, or empty list if none
        were found.
    """

    cards = await db_core.get_documents_by_property(
        DbCollection.CARDS, property_name, values, paginated, page, page_size
    )

    return [CardModelOut.model_validate(card) for card in cards]


async def delete_cards_by_id(card_ids: list[ObjectId]) -> int:
    """
    Deletes multiple cards from the database with the given IDs.

    Args:
        card_ids: The IDs of the cards to delete.

    Returns:
        The number of cards deleted, or None if none could be deleted.
    """

    delete_many_result = await db_core.delete_documents_by_id(DbCollection.CARDS, card_ids)
    if delete_many_result is not None:
        return delete_many_result.deleted_count


async def delete_cards_all() -> int:
    """
    Delete all cards in the database.

    Returns:
        The number of cards deleted, or None if none could be deleted.
    """

    delete_many_result = await db_core.delete_documents(DbCollection.CARDS)
    if delete_many_result is not None:
        return delete_many_result.deleted_count


# endregion



from collections import Counter
from sys import maxsize
from typing import Generic, Self

from scooze.card import CardT
from scooze.utils import ComparableObject, DictDiff


class DeckDiff(ComparableObject):
    """
    A class to represent a diff between two decks.

    Attributes:
        main: The diff between the main decks of two Decks.
        side: The diff between the sideboards of two Decks.
        cmdr: The diff between the command zones of two Decks.
    """

    def __init__(self, main: DictDiff, side: DictDiff, cmdr: DictDiff):
        self.main = main
        self.side = side
        self.cmdr = cmdr

    def __str__(self):
        if self.total() > 0:
            main_diff = str(self.main)
            side_diff = str(self.side)
            cmdr_diff = str(self.cmdr)
            return f"Main Diff:\n{main_diff}\nSide Diff:\n{side_diff}\nCmdr Diff:\n{cmdr_diff}"
        else:
            return ""

    def total(self) -> int:
        """
        The number of Cards in this DeckDiff.
        """

        return len(self.main) + len(self.side) + len(self.cmdr)


class DeckPart(ComparableObject, Generic[CardT]):
    """
    A class to represent a part of a deck.

    Attributes:
        cards: The cards in this DeckPart.
    """

    def __init__(self, cards: Counter[CardT] = None):
        self.cards = cards if cards is not None else Counter()

    def __getitem__(self, key: CardT):
        return self.cards[key]

    def __setitem__(self, key: CardT, value: int):
        self.cards[key] = value

    def __len__(self):
        return len(self.cards)

    def __str__(self):
        if len(self.cards) > 0:
            return "\n".join([f"{quantity} {card.name}" for card, quantity in self.cards.items()]) + "\n"
        else:
            return ""

    def total(self) -> int:
        """
        The number of cards in this DeckPart.
        """

        return self.cards.total()

    def diff(self, other: Self) -> DictDiff:
        """
        Generate a diff between this DeckPart and another.

        Args:
            other: The other DeckPart.

        Returns:
            A DictDiff with every card in both DeckParts and their counts.
        """

        return DictDiff.get_diff(self.cards, other.cards, NO_KEY=0)

    def add_card(self, card: CardT, quantity: int = 1) -> None:
        """
        Add a given quantity of a given card to this DeckPart.

        Args:
            card: The card to add.
            quantity: The number of copies of the card to be added.
        """

        self.cards.update({card: quantity})

    def add_cards(self, cards: Counter[CardT]) -> None:
        """
        Add the given cards to this DeckPart.

        Args:
            cards: The cards to add.
        """

        self.cards.update(cards)

    def remove_card(self, card: CardT, quantity: int = maxsize) -> None:
        """
        Remove a given quantity of a given card from this Deck. If quantity is
        not provided, removes all copies.

        Args:
            card: The card to remove.
            quantity: The number of copies of the card to be removed.
        """

        # using counterA - counterB results in a new Counter with only positive results
        self.cards = self.cards - Counter({card: quantity})

    def remove_cards(self, cards: Counter[CardT]) -> None:
        """
        Remove the given cards from this DeckPart.

        Args:
            cards: The cards to remove.
        """

        # using counterA - counterB results in a new Counter with only positive results
        self.cards = self.cards - cards



from cleo.commands.command import Command


class SetupLocalCommand(Command):
    name = "setup local"
    description = "Setup MongoDB locally."

    def handle(self):
        # TODO(#201) - Add local MongoDB built-in support
        self.line(
            "Usage: <fg=cyan>scooze setup docker</> or <fg=cyan>scooze setup local</>. Local support coming soon."
        )



import asyncio
import json
import os

import ijson
import scooze.database.deck as deck_db
from cleo.commands.command import Command
from cleo.helpers import option
from scooze.models.deck import DeckModelIn
from scooze.utils import DEFAULT_DECKS_DIR


class LoadDecksCommand(Command):
    name = "load decks"
    description = "Load sets of decks into the database."

    options = [
        option("all", description="All decks in the decks directory."),
        option("test", description="A set of Pioneer decks for testing purposes."),
        option(
            "decks-dir",
            description="Directory to store deck files, used with load-decks.",
            default=DEFAULT_DECKS_DIR,
            value_required=True,
            flag=False,
        ),
    ]

    def handle(self):
        # TODO(#145): Use ScoozeApi to load decks via API
        if self.option("all"):
            load_all_decks(self.option("decks-dir"))
        elif self.option("test"):
            load_test_decks()
        else:
            self.line("No files were selected to load.")


# TODO(#145): Can remove this once the command uses ScoozeApi
def load_all_decks(decks_dir: str):
    files = os.listdir(decks_dir)
    try:
        for file_path in files:
            with open(file_path, "r", encoding="utf8") as deck_file:
                print(f"Inserting {file_path.split('/')[-1]} file into the database...")
                decks = [
                    DeckModelIn.model_validate(deck_json)
                    for deck_json in ijson.items(
                        deck_file,
                        "item",
                    )
                ]
                asyncio.run(deck_db.add_decks(decks))
    except OSError as e:
        print(f"Encountered an error while trying to load {file_path.split('/')[-1]}")
        raise e


# TODO(#145): Can remove this once the command uses ScoozeApi
def load_test_decks():
    try:
        with open("./data/test/pioneer_decks.jsonl") as decks_file:
            print("Inserting test decks into the database...")
            json_list = list(decks_file)
            decks = [DeckModelIn.model_validate(json.loads(deck_json)) for deck_json in json_list]
            asyncio.run(deck_db.add_decks(decks))  # TODO(#7): this need async for now, replace with Python API
    except OSError as e:
        print("Encountered an error while trying to load test decks")
        raise e



from collections import Counter
from datetime import date

from pydantic import Field, model_validator
from scooze.catalogs import Format
from scooze.models.utils import ObjectIdT, ScoozeBaseModel
from scooze.utils import cmdr_size, main_size, side_size


class DeckModel(ScoozeBaseModel):
    """
    A model to represent a deck of Magic: the Gathering cards.

    Attributes:
        archetype: The archetype of this DeckModel.
        format: The format legality of the cards in this DeckModel.
        date_played: The date this DeckModel was played.
        main: The main deck. Typically 60 cards minimum.
        side: The sideboard. Typically 15 cards maximum.
        cmdr: The command zone. Typically 1 or 2 cards in Commander formats.
    """

    model_config = ScoozeBaseModel.model_config.copy()
    model_config["json_schema_extra"] = {
        "examples": [
            {
                "archetype": "Scooze Deck Example",
                "format": "Limited",
                "main": {
                    "6502bf99532dd43b31e6055a": 4,  # TODO(#6): replace with Python scooze id
                    "6502bf77bffae3b433093dcb": 4,  # TODO(#6): replace with Scavenging Ooze scooze id
                },
                "side": {
                    "6502bfe2e0e370d002c87ceb": 1,  # TODO(#6): replace with Keruga scooze id
                },
            },
        ]
    }

    archetype: str = Field(
        default="",
        description="The archetype of this Deck.",
    )
    format: Format = Field(
        default=Format.NONE,
        description="The format of the tournament where this Deck was played.",
    )
    date_played: date = Field(
        default=None,
        description="The date this Deck was played.",
    )
    main: Counter[ObjectIdT] = Field(
        default=Counter(),
        description="The main deck. Typically 60 cards minimum.",
    )
    side: Counter[ObjectIdT] = Field(
        default=Counter(),
        description="The sideboard. Typically 15 cards maximum.",
    )
    cmdr: Counter[ObjectIdT] = Field(
        default=Counter(),
        description="The command zone. Typically 1 or 2 cards in Commander formats.",
    )

    # region Validators

    # TODO(#49): add validation for fields

    @model_validator(mode="after")
    def validate_main(self):
        m_min, m_max = main_size(self.format)
        if self.main.total() < m_min:
            e = ValueError(
                f"Not enough cards in main deck. Provided main deck has {self.main.total()} cards. Expected at least {m_min}."
            )
            raise e
        elif self.main.total() > m_max:
            e = ValueError(
                f"Too many cards in main deck. Provided main deck has {self.main.total()} cards. Expected at most {m_max}."
            )
            raise e
        return self

    @model_validator(mode="after")
    def validate_side(self):
        s_min, s_max = side_size(self.format)
        if self.side.total() < s_min:
            raise ValueError(
                f"Not enough cards in sideboard. Provided sideboard has {self.side.total()} cards. Expected at least {s_min}."
            )
        elif self.side.total() > s_max:
            raise ValueError(
                f"Too many cards in sideboard. Provided sideboards has {self.side.total()} cards. Expected at most {s_max}."
            )
        return self

    @model_validator(mode="after")
    def validate_cmdr(self):
        c_min, c_max = cmdr_size(self.format)
        if self.cmdr.total() < c_min:
            raise ValueError(
                f"Not enough cards in command zone. Provided command zone has {self.cmdr.total()} cards. Expected at least {c_min}."
            )
        elif self.cmdr.total() > c_max:
            raise ValueError(
                f"Too many cards in command zone. Provided command zone has {self.cmdr.total()} cards. Expected at most {c_max}."
            )
        return self

    # endregion


class DeckModelIn(DeckModel):
    pass


class DeckModelOut(DeckModel):
    id: ObjectIdT = Field(
        default=None,
        alias="_id",
    )



import ijson
import scooze.database.card as db
from pydantic_core import ValidationError
from scooze.bulkdata import download_bulk_data_file_by_type
from scooze.catalogs import ScryfallBulkFile
from scooze.models.card import CardModelIn


async def load_card_file(file_type: ScryfallBulkFile, bulk_file_dir: str) -> None:
    """
    Loads the desired file from the given directory into a local Mongo
    database. Attempts to download it from Scryfall if it isn't found.

    Args:
        file_type: The type of [ScryfallBulkFile](https://scryfall.com/docs/api/bulk-data)
        to insert into the database.
        bulk_file_dir: The path to the folder containing the ScryfallBulkFile.
    """

    file_path = f"{bulk_file_dir}/{file_type}.json"
    try:
        with open(file_path, "rb") as cards_file:
            print(f"Loading {file_type} file into the database...")

            batch_size = 5000
            current_batch_count = 0
            results_count = 0
            current_batch = []
            card_jsons = ijson.items(cards_file, "item")

            async def load_batch(batch) -> int:
                batch_results = await db.add_cards(batch)
                if batch_results is not None:
                    return len(batch_results)
                return 0

            for card_json in card_jsons:
                if (validated_card := _try_validate_card(card_json)) is not None:
                    current_batch.append(validated_card)
                    current_batch_count += 1
                    if current_batch_count >= batch_size:
                        results_count += await load_batch(current_batch)
                        current_batch = []
                        current_batch_count = 0
                        print(f"Finished processing {results_count} cards...", end="\r")
            results_count += await load_batch(current_batch)

            print(f"Loaded {results_count} cards to the database.")

    except FileNotFoundError:
        print(file_path)
        download_now = input(f"{file_type} file not found; would you like to download it now? [y/N] ") in "yY"
        if not download_now:
            print("No cards loaded into database.")
            return
        download_bulk_data_file_by_type(file_type, bulk_file_dir)
        await load_card_file(file_type, bulk_file_dir)


def _try_validate_card(card_json: dict) -> CardModelIn | None:
    """
    Attempt to convert a single card's JSON to a model for DB import, and
    report validation errors that arise in conversion.

    Args:
        card_json: JSON representation of a single card object.

    Returns:
        A validated model, or None if validation failed.
    """

    try:
        card = CardModelIn.model_validate(card_json)
        return card

    except ValidationError as e:
        print(f"Card with name {card_json['name']} not added due to validation error: \n{e}")

        return



from typing import Any

import scooze.database.deck as db
from fastapi import APIRouter
from fastapi.responses import JSONResponse
from scooze.models.deck import DeckModelIn

router = APIRouter(
    prefix="/decks",
    tags=["decks"],
    responses={404: {"description": "Decks Not Found"}},
)


@router.get("/", summary="Get decks at random")
async def decks_root(limit: int = 3):
    """
    Get random decks up to the given limit.

    - **limit** - the maximum number of decks to get
    """

    decks = await db.get_decks_random(limit=limit)
    if decks is not None:
        return JSONResponse([deck.model_dump(mode="json") for deck in decks], status_code=200)
    else:
        return JSONResponse({"message": "No decks found in the database."}, status_code=404)


# Create


@router.post("/add", summary="Create new decks")
async def add_decks(decks: list[DeckModelIn]):
    inserted_ids = await db.add_decks(decks=decks)
    if inserted_ids is not None:
        return JSONResponse({"message": f"Created {len(inserted_ids)} deck(s)."}, status_code=200)
    else:
        return JSONResponse({"message": "Failed to create any new decks."}, status_code=400)


@router.post("/by", summary="Get decks by property")
async def get_decks_by(
    property_name: str, values: list[Any], paginated: bool = False, page: int = 1, page_size: int = 10
):
    """
    Get decks where the given property matches any of the given values.

    - **property_name** - the property to check against
    - **values** - matching values of the given property
    - **paginated** - return paginated results if True, return all matches if
    False
    - **page** - return matches from the given page
    - **page_size** - the number of results per page
    """

    decks = await db.get_decks_by_property(
        property_name=property_name, values=values, paginated=paginated, page=page, page_size=page_size
    )
    if decks is not None:
        return JSONResponse([deck.model_dump(mode="json") for deck in decks], status_code=200)
    else:
        return JSONResponse({"message": "Decks not found."}, status_code=404)


# Delete


@router.delete("/delete/all", summary="Delete all decks")
async def delete_decks_all():
    deleted_count = await db.delete_decks_all()

    if deleted_count is not None:
        return JSONResponse({"message": f"Deleted {deleted_count} deck(s)."}, status_code=200)
    else:
        return JSONResponse({"message": "No decks deleted."}, status_code=404)



import scooze.database.deck as deck_db
from cleo.commands.command import Command
from cleo.helpers import argument
from scooze.api import ScoozeApi
from scooze.catalogs import DbCollection

ACCEPTED_DELETE_ARGS = {
    "all",
    "cards",
    "decks",
}


class DeleteCommand(Command):
    name = "delete"
    description = "Delete collections from the database."

    arguments = [
        argument(
            "collections",
            f"Which collections to remove from the database. Can be any of: <fg=cyan>all, cards, decks</>",
            multiple=True,
        )
    ]

    def handle(self):
        to_delete: list[DbCollection] = []
        delete_args = set(self.argument("collections"))

        if "all" in delete_args:
            to_delete.extend(DbCollection.list())
        else:
            if "cards" in delete_args:
                to_delete.append(DbCollection.CARDS)
            if "decks" in delete_args:
                to_delete.append(DbCollection.DECKS)

        if len(to_delete) == 0:
            extra_args = " ".join(delete_args - ACCEPTED_DELETE_ARGS)
            self.line(f"No valid collections were given. Ignored the following: <fg=cyan>{extra_args}</>")

        for collection in to_delete:
            delete_collection(collection)


def delete_collection(coll: DbCollection):
    clean = input(f"Delete existing {coll}? [y/N] ") in "yY"
    if clean:
        print(f"Deleting all {coll} from your local database...")
        match coll:
            case DbCollection.CARDS:
                with ScoozeApi() as s:
                    s.delete_cards_all()
            case DbCollection.DECKS:
                # TODO(#145): Use the ScoozeApi for this
                deck_db.delete_decks_all()
